//Don't ask me why this needs to go up here, but it does.
#include <fstream>

#include <config.h>
#include <Console.h>
#include <model/Compiler.h>
#include <model/parser_extra.h>
#include <model/ParseTree.h>
#include <model/TraceMonitor.h>
#include <graph/NodeError.h>
#include <sampler/SamplerFactory.h>
#include <sampler/ConjugateFactory.h>
#include <sampler/FiniteFactory.h>
#include <sampler/SliceFactory.h>
#include <graph/Node.h>
#include <sarray/RightRangeIterator.h>
#include <sarray/Range.h>
#include <sarray/SArray.h>

#include "Rmath.h"

#include <map>
#include <list>
#include <stdexcept>
//#include <iostream>

#include <vector>

using std::ostream;
using std::ofstream;
using std::endl;
using std::flush;
using std::vector;
using std::string;
using std::map;
using std::list;

Console::Console(ostream &out, ostream &err)
  : _out(out), _err(err), _functab(), _disttab(), _symtab()
{
  _pvariables = 0;
  _prelations = 0;
  _model = 0;
}

Console::~Console()
{
  /* FIXME: When do we delete _pvariables, _prelations? */
  delete _model;
}

void Console::checkModel(string const &filename)
{
  FILE *file = fopen(filename.c_str(), "r");
  if (file) {
    if (_model) {
      _out << "Replacing existing model" << endl;
      delete _model;
    }
    _model = new Model;
    _out << "Parsing model file " << filename << endl;
    int lineno =  parse_bugs(file, &_pvariables, &_prelations);
    if (lineno != 0) {
      _out << "Parse error on line " << lineno << endl;
      delete _model;
      _model = 0;
    }
    fclose(file);
    /* FIXME: we should free the memory associated wth
       _pvariables, _prelations after compiling
       the model
    */
  }
  else {
    _out << "Unable to open file " << filename << endl;
  }
}

void Console::compile(map<string, SArray> const &data_table,
		      unsigned int nchain)
{
  if (nchain == 0) {
    _err << "You must have at least one chain" << endl;
  }
  if (nchain > 1) {
    _err << "Multiple chains not currently supported. Sorry" << endl;
  }
  if (_pvariables == 0) {
    _err << "No variables defined yet" << endl;
    return;
  }
  if (_prelations == 0) {
    _err << "No relations defined yet" << endl;
    return;
  }

  Compiler compiler(_model->graph(), _symtab, _functab, _disttab, 
		    data_table);

  try {
    _out << "Declaring variables" << endl;
    compiler.declareVariables(*_pvariables);
    _out << "Allocating nodes" << endl;
    compiler.allocateNodes(_prelations);
    compiler.fillConstants();
    _out << "Writing relations" << endl;
    compiler.writeRelations(_prelations);
    _out << "Collecting nodes" << endl;
    compiler.collectNodes();
    _out << "Writing data" << endl;
    compiler.writeValuesToGraph(data_table, true);
    _out << "Checking graph" << endl;
    _model->checkGraph();
    _out << "  Graph Size: " << _model->graph().size() << endl;
  }
  catch (NodeError except) {
    _err << "Error in node " << except.node->name() << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::logic_error except) {
    _err << "Logic error" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
}

static unsigned int asUnsignedInt(double y)
{
  if (y < 0) {
    throw std::runtime_error("Expected positive integer");  
  }
  else if (y > UINT_MAX) {
    throw std::runtime_error("Unable to coerce value to unsigned int: too large");
  }
  unsigned int i = static_cast<unsigned int>(y + DBL_EPSILON);
  if (fabs(y - i) > DBL_EPSILON) {
    throw std::runtime_error("Expected integer value");
  }
  return i;
}

void Console::initialize()
{
  if (_model == 0) {
    _err << "Can't initialize. No model!" << endl;
    return;
  }
  try {
    _model->initialize();

    vector<SamplerFactory const*> samplers;
    SamplerFactory *cfac = new ConjugateFactory;
    samplers.push_back(cfac);
    SamplerFactory *ffac = new FiniteFactory;
    samplers.push_back(ffac);
    SamplerFactory *sfac = new SliceFactory;
    samplers.push_back(sfac);
    _model->chooseSamplers(samplers);
    delete cfac;
    delete ffac;
    delete sfac;
  }
  catch (NodeError except) {
    _err << "Error in node " << except.node->name() << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::logic_error except) {
    _err << "Logic error" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
}

void Console::setParameters(map<string, SArray> const &init_table,
			    unsigned int chain)
{
  if (_model == 0) {
    _err << "Can't set initial values. No model!" << endl;    
    return;
  }
  if (chain > 1) {
    _err << "Multiple chains not currently supported. Sorry" << endl;
    return;
  }
  try {
    _symtab.writeData(init_table, false);
  }
  catch (NodeError except) {
    _err << "Error in node " << except.node->name() << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::logic_error except) {
    _err << "Logic error" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
}

void Console::update(long n)
{
  if (_model == 0) {
    _err << "Can't update. No model!" << endl;    
    return;
  }
  if (!_model->isInitialized()) {
    _err << "Model not initialized" << endl;
    return;
  }
  try {
    int nstar = 20;
    if (n >= nstar) {
      _out << "Updating: ";
       // Print a progress bar using stars
       long m = n/nstar;
       for (int i = 0; i < nstar - 1; i++) {
          _model->update(m);
          _out << "*" << flush;
       }
       _model->update(n - (nstar-1)*m);
       _out << "*" << endl << flush;
     }
     else {
       _model->update(n);
     }
  }
  catch (NodeError except) {
    _err << "Error in node " << except.node->name() << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
  catch (std::logic_error except) {
    _err << "Logic error" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
}

void Console::iter()
{
  if (!_model) {
    _err << "0" << endl;
  }
  else {
    _out << _model->iteration() << endl;
  }
}

void Console::setMonitor(string const &name, Range const &range,
			 unsigned int thin)
{
  if (!_model) {
    _out << "Can't set monitor. No model!" << endl;    
    return;
  }

  try {
    NodeArray *array = _symtab.getVariable(name);
    if (array) {
      if (array->range().contains(range)) {
	_model->setMonitor(array->getSubset(range), thin);
      }
      else {
	_err << "Requested invalid subset of node " << name << endl;
	return;
      }
    }
    else {
      _err << "Node " << name << " not found" << endl;
      return;
    }
  }
  catch (std::logic_error except) {
    _err << "Logic error:" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error:" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
}

void Console::clearMonitor(string const &name, Range const &range)
{
  if (!_model) {
    _out << "Can't clear monitor. No model!" << endl;    
    return;
  }

  try {
    NodeArray *array = _symtab.getVariable(name);
    if (array) {
      if (array->range().contains(range)) {
	Node const *node = array->get(range);
	if (node == 0) {
	  _err << "Node is not being ,monitored" << endl;
	  return;
	}
	else {
	  _model->clearMonitor(node);	  
	}
      }
      else {
	_err << "Requested invalid subset of node " << name << endl;
	return;
      }
    }
    else {
      _err << "Node " << name << " not found" << endl;
      return;
    }
  }
  catch (std::logic_error except) {
    _err << "Logic error:" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error:" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
}

void Console::setMonitor(string const &name, unsigned int thin)
{
  if (!_model) {
    _out << "Can't set monitor. No model!" << endl;    
    return;
  }

  NodeArray *array = _symtab.getVariable(name);
  if (array) {
    setMonitor(name, array->range(), thin);
  }
  else {
    _err << "Node " << name << " not found" << endl;
  }
}

void Console::clearMonitor(string const &name)
{
  if (!_model) {
    _out << "Can't clear monitor. No model!" << endl;    
    return;
  }

  try {
    NodeArray *array = _symtab.getVariable(name);
    if (array) {
      if (array->get(array->range())) {
	_model->clearMonitor(array->get(array->range()));	  
      }
      else {
	_err << "Node " << name << "is not being monitored" << endl;
      }
    }
    else {
      _err << "Node " << name << " not found" << endl;
      return;
    }
  }
  catch (std::logic_error except) {
    _err << "Logic error:" << endl;
    _err << except.what() << endl;
    _err << "Please send a bug report to " << PACKAGE_BUGREPORT << endl;
    errorRecovery();
  }
  catch (std::runtime_error except) {
    _err << "Runtime error:" << endl;
    _err << except.what() << endl;
    errorRecovery();
  }
}

static void writeDouble(double x, ostream &out)
{
  if (x == JAGS_NA) {
    out << "NA";
  }
  else if (ISNAN(x)) {
    out << "NaN";
  }
  else if (!R_FINITE(x)) {
    if (x > 0) {
      out << "Inf";
    }
    else {
      out << "-Inf";
    }
  }
  else {
    out << x;
  }
}

void Console::errorRecovery()
{
  _out << "Dumping state at iteration " << _model->iteration() 
       << " to file jags.dump" << endl;  
  ofstream out("jags.dump");
  if (!out) {
    _err << "Failed to open file \"jags.dump\" for dumping state" << endl;
    return;
  }
  else {
    dumpState(out, 1, false);
    delete _model; _model = 0;
  }
}  
 
/* FIXME: Something of a mish-mash here
   - The business of dumping to a file should be in the terminal
   not the library.  But we need return values on console
   functions before we can do this
*/

void Console::dumpState(ostream &out, unsigned int chain, bool observed)
{
  if (chain > 1) {
    _err << "Multiple chains not supported. Sorry" << endl;
  }
  map<string,SArray> data_table;
  _symtab.readData(data_table, observed);

  for (map<string, SArray>::const_iterator p = data_table.begin();
       p != data_table.end(); ++p) {
    string const &name = p->first;
    SArray const &sarray = p->second;
    double const *value = sarray.value();
    long length = sarray.length();
    out << "\"" << name << "\" <- " << endl;
    Index dim = sarray.range().dim(false);
    if (dim.length() == 1) {
      // Vector value
      if (dim[0] == 1) {
	writeDouble(value[0], out);
        out << endl;
      }
      else {
	out << "c(";
	for (int i = 0; i < length; ++i) {
	  if (i > 0) {
	    out << ",";
	  }
	  writeDouble(value[i], out);
	}
	out << ")" << endl;
      }
    }
    else {
      // Array value
      out << "structure(c(";
      for (int i = 0; i < length; ++i) {
	if (i > 0) {
	  out << ",";
	}
	writeDouble(value[i], out);
      }
      out << "), .Dim = c(";
      for (int j = 0; j < dim.length(); ++j) {
	if (j > 0) {
	  out << ",";
	}
	out << dim[j];
      }
      out << "))" << endl;
    }
  }
}  


void Console::coda(ostream &index, ostream &output)
{
  if (!_model) {
     _err << "Can't dump CODA output. No model!" << endl;
     return;
  }

  if (index == output) {
    _err << "Index and output streams for CODA are the same!" << endl;
    return;
  }

  vector<Node const *> nodes;
  list<TraceMonitor*> const &monitors = _model->monitors();
  for (list<TraceMonitor*>::const_iterator i = monitors.begin();
       i != monitors.end(); i++) 
    {
      nodes.push_back((*i)->node());
    }
  _model->coda(nodes, index, output);
}

void Console::coda(string const &name, Range const &range,
		   ostream &index, ostream &output)
{
  if (!_model) {
     _err << "Can't dump CODA output. No model!" << endl;
     return;
  }

  NodeArray *array = _symtab.getVariable(name);
  if (array) {
    if (array->range().contains(range)) {
      if (array->get(range)) {
	vector<Node const *> nodes(1, array->get(range));
	_model->coda(nodes, index, output);
      }
      else {
	_err << "Node is not being monitored" << endl;
      }
    }
    else {
      _err << "Requested invalid subset of node " << name << endl;
      return;
    }
  }
  else {
    _err << "Node " << name << " not found" << endl;
    return;
  }
}

void Console::coda(string const &name, ostream &index, ostream &output)
{
  if (!_model) {
    _err << "Can't dump CODA output. No model!" << endl;
    return;
  }
  
  NodeArray *array = _symtab.getVariable(name);
  if (array) {
    vector<Node const *> nodes(1, array->getSubset(array->range()));
    _model->coda(nodes, index, output);
  }
  else {
    _err << "Node " << name << " not found" << endl;
    return;
  }
}

Model const *Console::model()
{
  return _model;
}

void Console::setSeed(unsigned int seed)
{
  /* Given a single seed, calculates seeds for the Marsaglia
   * Multicarry random number generator
   */
  unsigned int mmseeds[2];
  for(int j = 0; j < 2; j++) {
    seed = (69069 * seed + 1);
    mmseeds[j] = seed;
  }
  /* Fix up zero seeds */
  if (mmseeds[0] == 0)
    mmseeds[0] = 1;
  if (mmseeds[1] == 0)
    mmseeds[1] = 1;

  set_seed(mmseeds[0], mmseeds[1]);
}
