#include <config.h>
#include <distributions/DBin.h>
#include <sarray/SArray.h>

#include <Rmath.h>

using std::vector;

#define ASLONG(x) (static_cast<long>(x + 1.0E-6))

DBin::DBin()
  : DistFinite("dbin", 2)
{}

DBin::~DBin()
{}

static void
getParameters(long &size, double &prob, vector<SArray const *> const &parameters)
{
  /* BUGS parameterization is in opposite order to R parameterization */

  size = ASLONG(*(parameters[1]->value()));
  prob = *(parameters[0]->value());
}

bool 
DBin::checkParameterDim (SArray const &value, 
			 vector<SArray const *> const &parameters) const
{
  return (parameters[0]->length() == 1) && (parameters[1]->length() == 1) 
    && (value.length() == 1);
}

bool DBin::checkParameterValue(vector<SArray const *> const &parameters) const
{
  long size = 0;
  double prob = 0;
  getParameters(size, prob, parameters);

  if (size < 1)
    return false;
  else if (prob < 0.0 || prob > 1.0)
    return false;
  else
    return true;
}

static double logdbinom(long x, long size, double prob)
{
  /* 
     Calculate log density without normalizing constant.
  */
  if (size == 0) {
    return x == 0 ? 0 : -DBL_MAX;
  }
  else if (prob < 0 || prob > 1 || x < 0 || x > size) {
    return -DBL_MAX;
  }
  else if (prob == 0) {
    return x == 0 ? 0 : -DBL_MAX;
  }
  else if (prob == 1) {
    return x == size ? 0 : -DBL_MAX;
  }
  else {
    return x * log(prob) + (size - x) * log(1 - prob);
  }
}

double DBin::logLikelihood(long x, vector<SArray const *> const &parameters) const
{
  double size = *(parameters[1]->value());
  double prob = *(parameters[0]->value());


  if (lowerBound(parameters) || upperBound(parameters)) {
    if (x < lowerLimit(parameters) || x > upperLimit(parameters)) {
      return -DBL_MAX;
    }
    else {
      // Renormalize
      double sump = pbinom(lowerLimit(parameters), size, prob, 1, 0) - 
	pbinom(upperLimit(parameters), size, prob, 1, 0);
      return dbinom(x, size, prob, 1) - log(sump);
    }
  }
  else {
    if (parameters[1]->isFixed()) {
      /* 
	 dbinom includes an expensive computation of the normalizing
	 constant. We try to get round this in the case where size
	 is fixed.
      */
      return logdbinom(x, ASLONG(size), prob);
    }
  }
}

long DBin::randomSample(vector<SArray const *> const &parameters) const
{
  long size = 0;
  double prob = 0.0;
  getParameters(size, prob, parameters);

  if (lowerBound(parameters) || upperBound(parameters)) {
    long lower = lowerLimit(parameters);
    long upper = upperLimit(parameters);
    double plower = pbinom(lower - 1, size, prob, 1, 0);
    double pupper = pbinom(upper, size, prob, 1, 0);
    double p = runif(plower, pupper);
    for (long i = lower; i < upper; i++) {
      if (pbinom(i, size, prob, 1, 0) > p)
	return i;
    }
    return upper;
  }
  else {
    return ASLONG(rbinom(size, prob));
  }
}

long DBin::lowerLimit(vector<SArray const *> const &parameters) const
{
  SArray const *lb = lowerBound(parameters);
  if (lb) {
    long lower = ASLONG(*(lb->value()));
    return lower > 0 ? lower : 0;
  }
  return 0;
}
	

long DBin::upperLimit(vector<SArray const *> const &parameters) const
{
  long size=0;
  double prob=0;
  getParameters(size, prob, parameters);

  SArray const *ub = upperBound(parameters);
  if (ub) {
    long upper = ASLONG(*(ub->value()));
    return upper < size ? upper : size;
  }
  return size;
}
