#include <config.h>
#include <distributions/DCat.h>
#include <sarray/SArray.h>

#include <Rmath.h>

#include <cfloat>
#include <algorithm>
#include <cmath>

using std::min;
using std::max;
using std::vector;

DCat::DCat()
  : DistFinite("dcat", 1) 
{}

DCat::~DCat()
{}

static void getParameters(double const **prob, long *length,
			  vector<SArray const *> const &parameters)
{
  if (prob != 0)
    *prob = parameters[0]->value();
  if (length != 0)
    *length = parameters[0]->length();
}

bool DCat::checkParameterDim(SArray const &value,
			     vector<SArray const *> const &parameters) const
{
  return (value.length() == 1) && (parameters[0]->length() >= 2);
}

bool DCat::checkParameterValue(vector<SArray const *> const &parameters) const
{
  double const *prob = 0;
  long length = 0;
  getParameters(&prob, &length, parameters);

  double sump = 0.0;
  for (long i = 0; i < length; i++) {
    if (prob[i] < 0.0 || prob[i] > 1.0) {
      return false; // Invalid probability
    }
    sump += prob[i];
  }
  if (fabs(sump - 1.0) > 1.0E-6)
    return false; // Probabilities do not sum to 1 

  return true;
}

long DCat::lowerLimit(vector<SArray const *> const &parameters) const
{
  if (lowerBound(parameters) != 0) {
    long lower = static_cast<long>(*(lowerBound(parameters)->value()));
    return max(lower, (long) 1);
  }
  else {
    return 1;
  }
}

long DCat::upperLimit(vector<SArray const *> const &parameters) const
{
  long length = 0;
  getParameters(0, &length, parameters);

  if (upperBound(parameters) != 0) {
    long ub = static_cast<long>(*(upperBound(parameters)->value()) + 1.0E-6);
    return min(ub, length);
  }
  else {
    return length;
  }
}

double DCat::logLikelihood(long x, vector<SArray const *> const &parameters) const
{
  long lower = lowerLimit(parameters);
  long upper = upperLimit(parameters);

  if (x < lower || x > upper) {
    return -DBL_MAX;
  }
  else {
    double const *prob = 0;
    getParameters(&prob, 0, parameters);

    if (lowerBound(parameters) || upperBound(parameters)) {
      /* Bounded: need to renormalize */
      double sump = 0.0;
      for (long i = lower; i <= upper; i++) {
	sump += prob[i - 1];
      }
      return prob[x-1] == 0 ? -DBL_MAX : log(prob[x - 1]) - log(sump);
    }
    else {
      return prob[x-1] == 0 ? -DBL_MAX : log(prob[x - 1]);
    }
  }
}

long DCat::randomSample(vector<SArray const *> const &parameters) const
{
  long lower = lowerLimit(parameters);
  long upper = upperLimit(parameters);
  
  double const *prob = 0;
  getParameters(&prob, 0, parameters);

  double sump = 1.0;
  if (lowerBound(parameters) || upperBound(parameters)) {
    /* Renormalize */
    for (long i = lower; i <= upper; i++) {
      sump += prob[i - 1];
    }
  }

  /* Draw the sample */
  double q = runif(0, sump);

  /* Categorize */
  sump = 0;
  for (long j = lower; j < upper; j++) {
    sump += prob[j - 1];
    if (q <= sump) {
      return j;
    }
  }
  return upper;
}
