#include <config.h>
#include <distributions/DDirch.h>
#include <sarray/SArray.h>

#include <Rmath.h>

#include <stdexcept>
#include <cmath>

using std::vector;
using std::length_error;

/* The Dirichlet distribution is extended to allow zero shape parameters.
   These represent structural zeros: when x ~ ddirch(alpha) is forward
   sampled, x[i] = 0 when alpha[i] = 0. To avoid trapping states in the
   model, structural zeros are only allowed when the alpha parameter
   is fixed.
*/

DDirch::DDirch()
  : Distribution("ddirch", 1) 
{}

DDirch::~DDirch()
{}

static void getParameters (double const * &alpha, long &length,
			   vector<SArray const *> const &parameters)
{
  alpha = parameters[0]->value();
  length = parameters[0]->length();
}

bool DDirch::checkParameterDim(SArray const &value,
			       vector<SArray const *> const &parameters) const
{
  return (parameters[0]->length() == value.length());
}

bool DDirch::checkParameterValue(vector<SArray const *> const &parameters) const
{
  double const *alpha = 0;
  long length = 0;
  getParameters (alpha, length, parameters);

  int nzero = 0; // No of zero shape parameters
  for (int i = 0; i < length; i++) {
    if (alpha[i] <= 0) {
      if (alpha[i] == 0) {
	nzero++;
      }
      else {
	return false;
      }
    }
  }

  if (nzero) {
    return parameters[0]->isFixed() && (nzero < length);
  }
  else {
    return true;
  }
}

double DDirch::logLikelihood(SArray const &x,
			     vector<SArray const *> const &parameters) const
{
  double const *y = x.value();
  double const *alpha = 0;
  long length = 0;
  getParameters(alpha, length, parameters);

  double alphasum = 0.0;
  double loglik = 0.0;
  for (long i = 0; i < length; i++) {
    if (alpha[i] == 0) {
      if (y[i] > 0)
	return -DBL_MAX;
    }
    else {
      alphasum += alpha[i];
      loglik += (alpha[i] - 1) * y[i] - lgamma(alpha[i]);
    }
  }
  loglik += lgamma(alphasum);

  return loglik;
}

void DDirch::randomSample(SArray &x,
			  vector<SArray const *> const &parameters) const
{
  double const *alpha = 0;
  long length = 0;
  getParameters(alpha, length, parameters);

  if (length != x.length()) 
    throw length_error ("Length mismatch error in DDirch::randomSample");

  /* Generate independent gamma random variables, then normalize
     to create Dirichlet distribution.
  */
  double *y = new double[length];
  double sumy = 0.0;
  for (long i = 0; i < length; i++) {
    y[i] = (alpha[i]==0) ? 0 : rgamma(alpha[i], 1);
    sumy += y[i];
  }
  for (long j = 0; j < length; j++) {
    y[j] /= sumy;
  }
  x.setValue(y, length);
  delete [] y;
}

