#include <config.h>
#include <distributions/DMState.h>
#include <sarray/SArray.h>
#include <matrix/matexp.h>

#include <Rmath.h>

#include <cfloat>
#include <algorithm>
#include <cmath>

using std::min;
using std::max;
using std::vector;

//debuggin
#include <iostream>
using std::cout;
using std::endl;

DMState::DMState()
  : DistFinite("dmstate", 3) 
{}

DMState::~DMState()
{}

static void getParameters(int *initial, double *time,
			  int *nstates, double const **intensity, 
			  vector<SArray const *> const &parameters)
{
  *initial = static_cast<int>(*(parameters[0]->value()) + 1.0E-6);
  *time = *(parameters[1]->value());
  *nstates = parameters[2]->dim(true)[0];
  *intensity = parameters[2]->value();
}

bool DMState::checkParameterDim(SArray const &value,
				vector<SArray const *> const &parameters) const
{
  if (value.length() != 1) // Value is scalar
    return false;

  if (parameters[0]->length() != 1) // Initial state is scalar
    return false;
  else if (parameters[1]->length() != 1) // Time is scalar
    return false;
  else {
    Index dim = parameters[2]->dim(true); // Intensity is a square matrix
    return (dim.length() == 2) && (dim[0] == dim[1]);
  }
}

bool DMState::checkParameterValue(vector<SArray const *> const &parameters) const
{
  int initial = 0;
  double time = 0;
  int nstates = 0;
  double const *intensity;
  getParameters(&initial, &time, &nstates, &intensity, parameters);

  // Number of states must be greater than 1
  if (nstates <= 1)
    return false;

  // Current state must be in range
  if (initial < 1 || initial > nstates)
    return false;

  // Time must be positive
  if (time < 0)
    return false;

  // Intensity parameter must be a real intensity matrix
  for (int i = 0; i < nstates; ++i) {
    double sum = 0.0;
    for (int j = 0; j < nstates; ++j) {
      double rate = intensity[i + nstates * j];
      // Diagonal elements are non-positive. Others are non-negative
      if (i == j) {
	if (rate > 0.0)
	  return false;
      }
      else {
	if (rate < 0.0)
	  return false;
      }
      sum += rate;
    }
    // Rows of intensity matrix sum to 0
    if (fabs(sum) > 1.0E-6) {
      // fixme: we should relate the tolerance to the machine precision
      return false;
    }
  }

  return true;
}

long DMState::lowerLimit(vector<SArray const *> const &parameters) const
{
  // Fixme: ignoring lower and upper bounds
  return 1;
}

long DMState::upperLimit(vector<SArray const *> const &parameters) const
{
  // Fixme: ignoring lower and upper bounds
  return (parameters[2]->dim(true))[0];
}

double DMState::logLikelihood(long x, vector<SArray const *> const &parameters) const
{
  int initial = 0;
  double time = 0;
  int nstate = 0;
  double const *intensity = 0;
  getParameters(&initial, &time, &nstate, &intensity, parameters);
  
  if (x < 1 || x > nstate) {
    return -DBL_MAX;
  }
  else {
    /*
      FIXME: There is a big problem here with hidden Markov models.
      If the node is not observed, then we will need to evaluate the
      log-likelihood at different values of x, but with the same
      parameters.  This is hugely inefficient at the moment because 
      we calculate - and THROW AWAY - the transition probability matrix
      each time.  

      SOLUTION: Maybe a cache?
    */
    double *P = new double[nstate*nstate];
    //MatrixExp(P, intensity, nstate, time);
    MatrixExpPade(P, intensity, nstate, time);
    double lik = P[(initial - 1) + nstate * (x - 1)];
    delete [] P;
    if (lik <= 0) {
      /*
	Allow for some numerical imprecision that may create small
	negative probabilities
      */
      return -DBL_MAX;
    }
    else {
      return log(lik);
    }
  }
}

long DMState::randomSample(vector<SArray const *> const &parameters) const
{
  int initial = 0;
  double time = 0;
  int nstate = 0;
  double const *intensity;
  getParameters(&initial, &time, &nstate, &intensity, parameters);
  
  double *P = new double[nstate*nstate];
  //MatrixExp(P, intensity, nstate, time);
  MatrixExpPade(P, intensity, nstate, time);

  /* Draw the sample */
  double q = runif(0.0, 1.0);

  /* Categorize */
  double sump = 0.0;
  for (long j = 1; j < nstate; j++) {
    sump += P[(initial - 1) + nstate * (j - 1)];
    if (q <= sump) {
      delete [] P;
      return j;
    }
  }

  delete [] P;
  return nstate;
}
