#include <config.h>
#include <distributions/DMulti.h>
#include <sarray/SArray.h>

#include <Rmath.h>

#include <stdexcept>
#include <cmath>

using std::vector;
using std::length_error;

DMulti::DMulti()
  : Distribution("dmulti", 2) 
{}

DMulti::~DMulti()
{}

static void getParameters(double const * &prob, long &length, long &N,
			  vector<SArray const *> const &parameters)
{
  prob = parameters[0]->value();
  length = parameters[0]->length();
  N = (long) (*(parameters[1]->value()) + 0.5);
}

bool DMulti::checkParameterDim(SArray const &value,
			       vector<SArray const *> const &parameters) const
{
  if (value.dim(true).length() != 1)
    return false;
  if (parameters[0]->dim(true).length() != 1)
    return false;
  if (parameters[0]->length() != value.length())
    return false;
  if (parameters[1]->length() != 1)
    return false;
  return true;
}

bool DMulti::checkParameterValue(vector<SArray const *> const &parameters) const
{
  double const *prob = 0;
  long length = 0, N=0;
  
  getParameters(prob, length, N, parameters);
  if (N < 1)
    return false;

  double sump = 0.0;
  for (int i = 0; i < length; i++) {
    if (prob[i] < 0 || prob[i] > 1)
      return false;
    sump += prob[i];
  }
  if (fabs(sump - 1.0) > 1.0E-6)
    return false;

  return true;
}

double DMulti::logLikelihood(SArray const &x,
			     vector<SArray const *> const &parameters) const
{
  double const *y = x.value();
  double const *prob = 0;
  long N = 0;
  long length = 0;
  getParameters(prob, length, N, parameters);

  double loglik = 0.0;
  double ysum = 0.0;
  for (long i = 0; i < length; i++) {
    if (prob[i] == 0) {
      if (y[i] != 0)
	return -DBL_MAX;
    }
    else {
      loglik += y[i] * log(prob[i]) - lgamma(y[i] + 1);
      ysum += y[i];
    }
  }
  loglik += lgamma(ysum + 1);

  return loglik;
}

void DMulti::randomSample(SArray &x,
			  vector<SArray const *> const &parameters) const
{
  double const *prob = 0;
  long N = 0;
  long length = 0;
  getParameters(prob, length, N, parameters);
  
  if (x.length() != length)
    throw length_error ("Length mismatch in DMulti::randomSample");

  /* Sample multinomial as a series of binomial distributions */
  double *y = new double[length];
  double sump = 1.0;
  for (long i = 0; i < length - 1; i++) {
    y[i] = rbinom(N, prob[i]/sump);
    N -= (long) (y[i] + 0.5);
    if (N == 0)
      break;
    sump -= prob[i];
  }
  y[length - 1] = N;
  x.setValue(y, length);
  delete [] y;
}

