#include <config.h>
#include <distributions/DWish.h>
#include <matrix/lapack.h>
#include <matrix/matrix.h>
#include <sarray/SArray.h>

#include <Rmath.h>

#include <stdexcept>
#include <cfloat>
#include <cmath>
#include <vector>

using std::vector;
using std::logic_error;
using std::runtime_error;

DWish::DWish()
  : Distribution("dwish", 2) 
{}

DWish::~DWish()
{}

double DWish::logLikelihood(SArray const &x,
			    vector<SArray const *> const &parameters) const
{
  double const *y = x.value();
  double const *R = parameters[0]->value();
  double k = *parameters[1]->value();
  long p = parameters[0]->dim(true)[0];

  double loglik = 0;
  long N = p*p;
  for (int i = 0; i < N; ++i) {
    loglik += R[i] * y[i];
  }
  loglik += k * logdet(R, p) + (k - p - 1) * logdet(y, p);
  loglik /= 2;
  return loglik;
}


void DWish::randomSample(SArray &x,
			 vector<SArray const *> const &parameters) const
{
  double const *R = parameters[0]->value();
  double k = *parameters[1]->value();
  int nrow = parameters[0]->dim(true)[0];

  long length = x.length();
  double *y = new double[length];
  randomSample(y, length, R, k, nrow);
  x.setValue(y, length);
  delete [] y;
}

void DWish::randomSample(double *x, long length,
			 double const *R, double k, int nrow)
{
  /* 
     Generate random Wishart variable, using an algorithm proposed
     by Bill Venables and originally implemented in S.
  */
  if (length != nrow*nrow) {
    throw logic_error("invalid length in DWish::randomSample");
  }

  /* 
     Get inverse of R. Venables' algorithm was implemented in
     terms of the inverse of R, but we use a different parameterization
     to preserve conjugacy.
  */
  double * C = new double[length];
  inverse(C, R, nrow, true);

  /* Get Choleskly decomposition of C */
  int info = 0;
  F77_DPOTRF("U", &nrow, C, &nrow, &info);
  if (info != 0) {
    throw runtime_error("Failed to get Cholesky decomposition of R in dwish");
  }

  /* Set lower triangle of C to zero */
  for (int j = 0; j < nrow; j++) {
    double * C_j = &C[j*nrow]; //column j of matrix C
    for (int i = j + 1; i < nrow; i++) {
      C_j[i] = 0;
    }
  }

  /* Generate square root of Wishart random variable:
     - diagonal elements are square root of Chi square
     - upper off-diagonal elements are normal
     - lower off-diagonal elements are zero
  */
  double *Z = new double[length];
  for (int j = 0; j < nrow; j++) {
    double *Z_j = &Z[j*nrow]; //jth column of Z
    for (int i = 0; i < j; i++) {
      Z_j[i] = rnorm(0,1);
    }
    Z_j[j] = sqrt(rchisq(k - j));    
    for (int i = j + 1; i < nrow; i++) {
      Z_j[i] = 0;
    }
  }
  
  /* Transform Z with Cholesky decomposition */
  double *Ztrans = new double[length];
  for (int i = 0; i < nrow; i++) {
    for (int j = 0; j < nrow; j++) {
      double zz = 0;
      for (int l = 0; l < nrow; l++) {
	zz += Z[nrow * l + i] * C[nrow * j + l];
      }
      Ztrans[nrow * j + i] = zz;
    }
  }
  delete [] C;
  delete [] Z;

  /* Now put cross-product into x */
  for (int i = 0; i < nrow; i++) {
    double const *Ztrans_i = &Ztrans[nrow * i];
    for (int j = 0; j <= i; j++) {
      double const *Ztrans_j = &Ztrans[nrow * j];
      double xx = 0;
      for (int l = 0; l < nrow; l++) {
	xx += Ztrans_i[l] * Ztrans_j[l];
      }
      x[nrow * j + i] = x[nrow * i + j] = xx;
    }
  }
  delete [] Ztrans;
}

bool DWish::checkParameterDim (SArray const &value,
			       vector<SArray const *> const &parameters) const
{
  Index const &dim1 = value.dim(true);

  if (dim1.length() != 2)
    return false;
  if (dim1[0] != dim1[1])
    return false;
  if (parameters[0]->dim(true) != dim1)
    return false;
  if (parameters[1]->length() != 1)
    return false;

  return true;

}

bool DWish::checkParameterValue (vector<SArray const *> const &parameters) const
{
  /* Check that we have sufficient degrees of freedom */
  int nrow = parameters[0]->dim(true)[0];
  double k = *parameters[1]->value();
  if (k < nrow)
    return false;
  /* Check symmetry of mean*/
  double const *R = parameters[0]->value();
  for (int i = 0; i < nrow; ++i) {
    for (int j = 0; j < i; ++i) {
      if (fabs(R[i + nrow*j] - R[j + nrow*i]) > DBL_EPSILON)
	return false;
    }
  }
  /* FIXME? Skipping check of positive definiteness */
  return true;
}
