#ifndef DIST_FINITE_H_
#define DIST_FINITE_H_

#include <distributions/Distribution.h>

/** 
 * Abstract class for distributions that take a finite number of integer
 * values.
 * @short Finite categorical distributions
 */
class DistFinite : public Distribution
{
public:
  /**
   * Constructor
   * @see Distribution::Distribution
   */
  DistFinite(std::string const &name, unsigned int npar);
  /**
   * @returns the smallest integer value that a variable
   * may take with this distribution.
   */
  virtual long lowerLimit(std::vector<SArray const *> const &parameters) const = 0;
  /**
   * @returns the largest integer value that a variable may
   * take with this distribution.
   */
  virtual long upperLimit(std::vector<SArray const *> const &parameters) const = 0;
  /**
   * Implementation of the virtual member function
   * Distribution::logLikelihood. The implementation uses the
   * overloaded function logLikelihood(long,  SArray const**)
   * @see Distribution#logLikelihood
   */
  double logLikelihood(SArray const &x,
		       std::vector<SArray const *> const &parameters) const;
  /**
   * Implementation of the virtual member function
   * Distribution::randomSample. The implementation uses the
   * overloaded function randomSample(SArray const**)
   * @see Distribution#randomSample
   */
  void randomSample(SArray &x,
		    std::vector<SArray const *> const &parameters) const;
  /** 
   * Convenience function to make it easier for subclasses of
   * DistFinite to implement the logLikelihood member function.
   *
   * This function takes a long argument instead of an array
   * of doubles
   */
  virtual double logLikelihood(long x, 
			       std::vector<SArray const *> const &parameters) const = 0;
  /** 
   * Convenience function to make it easier for subclasses of
   * DistFinite to implement the randomSample member function.
   *
   * This function returns a long instead of writing the sampled
   * value to an array of doubles.
   */
  virtual long randomSample(std::vector<SArray const *> const &parameters) const  = 0;
};

DistFinite const *asFinite(Distribution const *dist);

#endif /* DIST_FINITE_H_ */
