#include <config.h>
#include <distributions/DistReal.h>
#include <sarray/SArray.h>

#include <Rmath.h>

#include <stdexcept>
#include <cfloat>
#include <cmath>
#include <algorithm>

using std::max;
using std::min;
using std::string;
using std::vector;
using std::length_error;

DistReal::DistReal(string const &name, unsigned int npar,
			       Support support, int lower, int upper)

  : Distribution(name, npar), _support(support), _lower(lower), _upper(upper)
{
}

double DistReal::logLikelihood(SArray const &x,
			       vector<SArray const *> const &parameters) const
{
  /* FIXME: This code is identical to DistFinite::logLikelihood */

  double y = x.value()[0];

  bool bb = lowerBound(parameters) != 0;
  bool ba = upperBound(parameters) != 0;

  double lower, upper;
  if (bb || ba) {
    lower = lowerLimit(parameters);
    upper = upperLimit(parameters);    
    if (y < lower || y > upper) {
      return -DBL_MAX;
    }
  }

  double loglik = d(y, parameters, true);
  if (bb && ba) {
    loglik -= log(p(upper, parameters, true, false) - 
		  p(lower, parameters, true, false));
  }
  else if (bb) {
    loglik -= p(lower, parameters, false, true);
  }
  else if (ba) {
    loglik -= p(upper, parameters, true, true);
  }
  return loglik;
}

void DistReal::randomSample(SArray &x,
			    vector<SArray const *> const &parameters) const
{
  if (x.length() != 1)
    throw length_error("length mismatch in DistReal::randomSample");

  bool bb = lowerBound(parameters) != 0;
  bool ba = upperBound(parameters) != 0;
  
  double y;
  if (!ba && !bb) {
    y = r(parameters);
  }
  else if (bb && ba) {
    double plower = p(lowerLimit(parameters), parameters, false, false);
    double pupper = p(upperLimit(parameters), parameters, false, false);
    double u = runif(plower, pupper);
    y = q(u, parameters, true, false);
  }
  else if (bb) {
    double logplower = p(lowerLimit(parameters), parameters, false, true);
    double logpx = logplower - rexp(1);
    y = q(logpx, parameters, false, true);
  }
  else if (ba) {
    double logplower = p(upperLimit(parameters), parameters, true, true);
    double logpx = logplower - rexp(1);
    y = q(logpx, parameters, true, true);
  }
  x.setValue(&y,1);
}

static double lowerDefault(Support support)
{
  double lower;

  switch(support) {
  case DIST_REAL:
    lower =  -DBL_MAX;
    break;
  case DIST_POSITIVE_REAL: case DIST_PROPORTION:
    lower =  0;
    break;
  }
  return lower;
}

static double upperDefault(Support support)
{
  double upper;

  switch(support) {
  case DIST_REAL: case DIST_POSITIVE_REAL:
    upper = DBL_MAX;
    break;
  case DIST_PROPORTION:
    upper = 1;
    break;
  }

  return upper;
}

double DistReal::lowerLimit(vector<SArray const *> const &parameters) const
{
  double ll = lowerDefault(_support);
  if (_lower != 0)
    ll = max(ll, *(parameters[_lower - 1]->value()));
  SArray const *pl = lowerBound(parameters);
  if (pl) {
    return max(ll, *(pl->value()));
  }
  else
    return ll;
}

double DistReal::upperLimit(vector<SArray const *> const &parameters) const
{
  double ul = upperDefault(_support);
  if (_upper != 0)
    ul = min(ul, *(parameters[_upper - 1]->value()));
  SArray const *pu = upperBound(parameters);
  if (pu)
    return min(ul, *(pu->value()));
  else
    return ul;
}

DistReal const *asReal(Distribution const *dist) {
	return dynamic_cast<DistReal const*>(dist);
}

bool 
DistReal::checkParameterDim (SArray const &value,
			     vector<SArray const *> const &parameters) const
{
  if (value.length() != 1)
    return false;

  for (unsigned int i = 0; i < parameters.size(); i++) {
    if (parameters[0]->length() != 1)
      return false;
  }
  return true;
}
