#ifndef LOGICAL_NODE_H_
#define LOGICAL_NODE_H_

#include <graph/DeterministicNode.h>
#include <vector>
class Function;
class SArray;

/**
 * The value of a logical node is a function of its parents.
 *
 * @short Logical Node
 */
class LogicalNode : public DeterministicNode {
    Function const *_func;
    std::vector<SArray const *> _parameters;

    /* Forbid copying of logical nodes */
    LogicalNode(LogicalNode const &orig);
    LogicalNode &operator=(LogicalNode const &rhs);

 public:
    //LogicalNode();
    /** Constructor. The function and parameters are set later  */
    LogicalNode(Index const &dim);
    ~LogicalNode();
    
    /** Sets the function of the node to func */
    void setFunction(Function const *func);
    /** Sets the parameters of the node. The parameters are added as
     * parents. It is an error to attempt to set the parameters before
     * the function.
     *
     * This function conducts checks that the length of the parameter
     * vector is correct for the function, and then checks that the
     * dimensions of the parameters are conforming.  It will throw
     * an exception if either check fails.
     *
     * @exception NodeError 
     * @see Function#checkDim 
     */
    void setParameters(std::vector<Node*> const &parameters);
    /** Returns the parameters of the node as a vector of SArray pointers  */
    std::vector<SArray const *> const &parameters() const;
    /** Returns the Function of the logical node */
    Function const* function() const;
    /** Calculates the value of the node based on the parameters. This
	function calls Function::checkParameterValue to 
     */
    void forwardSample();
    /**
     * Initializes node.
     *
     * This function will throw an exception if either the function
     * or the parameters have not been set.
     *
     * @exception NodeError
     */
    void initialize();
};

bool isLogical(Node const *node);
LogicalNode *asLogical(Node *node);

#endif /* LOGICAL_NODE_H_ */
