#include <config.h>
#include <graph/Graph.h>
#include <graph/StochasticNode.h>
#include <graph/NodeError.h>
#include <distributions/Distribution.h>
#include <sarray/Index.h>

#include <list>
#include <vector>
#include <string>
#include <stdexcept>

using std::list;
using std::vector;
using std::string;
using std::logic_error;

StochasticNode::StochasticNode(Index const &dim)
  : Node(dim), _dist(0), _parameters()
{
}

StochasticNode::~StochasticNode()
{
}

bool StochasticNode::isBounded() const
{
  if (_parameters.empty()) {
    return false;
  }
  else {
    return _dist->lowerBound(_parameters) || _dist->upperBound(_parameters);
  }
}

SArray const *StochasticNode::lowerBound()
{
  return _dist->lowerBound(_parameters);
}

SArray const *StochasticNode::upperBound()
{
  return _dist->upperBound(_parameters);
}

void StochasticNode::setBounds(Node *lBound, Node *uBound)
{
  Index dim = data.dim(true);

  if (_dist == 0) {
    throw NodeError(this,"Cannot set bounds before choosing distribution");
  }
  /*
  else if (_parameters.size() != _dist->npar()) {
    throw NodeError(this,"Cannot set bounds before parameters");
  }
  */
  if (_parameters.empty()) {
    throw NodeError(this, "Cannot set bounds before parameters");
  }
  if ((lBound && lBound->data.dim(true) != dim) || 
      (uBound && uBound->data.dim(true) != dim)) 
    {
      throw NodeError(this,"Dimension mismatch when setting bounds");
    }

  if (lBound) {
    SArray const *lb = _parameters[_parameters.size() - 2];
    if (lb != 0) {
      throw NodeError(this, "Attempt to reset lower bound");
    }
    _parameters[_parameters.size() - 2] = &lBound->data;
    addParent(lBound);      
  }
  if (uBound) {
    SArray const *ub = _parameters[_parameters.size() - 1];
    if (ub != 0) {
      throw NodeError(this, "Attempt to reset upper bound");
    }
    _parameters[_parameters.size() - 1] = &uBound->data;
    addParent(uBound);
  }
}


Distribution const *StochasticNode::distribution() const
{
    return _dist;
}

void StochasticNode::setDistribution(Distribution const *dist)
{
  if (_dist != 0) {
    string msg("Distribution already set for node ");
    msg.append(name());
    throw logic_error(msg);
  }

  _dist = dist;
  /* Now we know the distribution, we can set the parameter vector
     to the right size. Just push back zeros 
  _parameters.reserve(dist->npar() + 2);
  */
}

void StochasticNode::setParameters(vector<Node *> const &parameters)
{
  if (_dist == 0) {
    throw NodeError(this, "Can't set parameters of stochastic node: distribution not set");
  }
  if (_parameters.size() != 0) {
    throw NodeError(this, "Attempt to reset parameters");
  }

  for (unsigned int i = 0; i < parameters.size(); ++i) {
    _parameters.push_back(&(parameters[i]->data));
    this->addParent(parameters[i]);
  }
  // Push back zeros for the upper and lower bounds
  _parameters.push_back(0);
  _parameters.push_back(0);
  
  if (!_dist->checkParameterLength(_parameters)) {
    throw NodeError(this, "Incorrect number of parameters for distribution");
  }
  if (!_dist->checkParameterDim(data, _parameters)) {
    throw NodeError(this,"Invalid parameter dimensions for distribution");
  }
}

double StochasticNode::logDensity()
{
  if (_dist == 0)
    throw NodeError(this, "Distribution not set");

  return _dist->logLikelihood(data, _parameters);
}

vector<SArray const *> const &StochasticNode::parameters() const
{
  return _parameters;
}

void StochasticNode::forwardSample()
{
  _dist->randomSample(data, _parameters);
}  

StochasticNode *asStochastic(Node *node)
{
  return dynamic_cast<StochasticNode*>(node);
}

bool isStochastic(Node const *node)
{
  return dynamic_cast<StochasticNode const*>(node);
}



