#include <config.h>
#include <graph/SubSetIndex.h>
#include <graph/Node.h>

#include <stdexcept>
#include <string>
#include <vector>
#include <cfloat>
#include <iostream>

/* FIXME: Should be throwing exception instead of writing to cout */
using std::cout;
using std::endl;

using std::logic_error;
using std::string;
using std::vector;

SubSetIndex::SubSetIndex()
{
  _value = 1;
  _node = 0;
}

SubSetIndex::SubSetIndex(long value)
{
  _value = value;
  _node = 0;
}

SubSetIndex::SubSetIndex(Node *node)
{
  _value = 0;
  _node = node;
  if (node->data.length() != 1) {
    string msg("Node ");
    msg.append(node->name());
    msg.append("in SubsetIndex constructor is not scalar");
    throw logic_error(msg);
  }
}

static long asInteger(double x)
{
  return static_cast<long>(x + DBL_EPSILON);
} 

long SubSetIndex::value() const
{
  if (_node) {
    return asInteger(*(_node->data.value()));
  }
  else
    return _value;
}

Node *SubSetIndex::node() const
{
  return _node;
}

bool SubSetIndex::checkLimits(vector<SubSetIndex> const &lower, 
			      vector<SubSetIndex> const &upper)
{
  /* Check that the upper and lower limits define a subset
     that is of fixed dimension
  */
  unsigned int ndim = lower.size();
  if (ndim != upper.size())
    return false;

  for (unsigned int i = 0; i < ndim; i++) {
    Node const *lnode = lower[i].node();
    if (lnode != 0) {
      /* Lower bound is defined by a node, so it must be
	 1) Scalar
	 2) Equal to the upper bound 
      */
      if (lnode->data.length() != 1) {
	cout << "Parameter " << i << " not scalar" << endl;
	return false;
      }
      if (upper[i].node() != lnode) {
	cout << "Parameter" << i << " is node but UB != LB" << endl;
	return false;
      }
    }
    else {
      /* Lower bound is a fixed value, so the upper bound must be
	 1) Fixed
	 2) Greater than the lower bound
      */
      if (upper[i].node() != 0) {
	cout << "Parameter " << i << " LB is fixed but UB is scalar" << endl;
	return false;
      }
      if (upper[i].value() < lower[i].value()) {
	cout << "Parameter " << i << " LB = " << lower[i].value() <<
	  ", UB = " << upper[i].value() << endl;
	return false;
      }
    }
  }
  return true;
}

Index SubSetIndex::dim(vector<SubSetIndex> const &lower, 
		       vector<SubSetIndex> const &upper)
{
  if (!checkLimits(lower, upper)) {
    throw logic_error("Invalid limits for SubSetNode");
  }

  unsigned int ndim = lower.size();

  Index dim(ndim);
  for (unsigned int i = 0; i < ndim; i++) {
    if (lower[i].node() != 0) {
      dim[i] = 1;
    }
    else {
      dim[i] = upper[i].value() - lower[i].value() + 1;
    }
  }
  return dim;
}
