#include <config.h>
#include <graph/SubSetNode.h>
#include <graph/Node.h>
#include <sarray/LeftRangeIterator.h>

#include <stdexcept>
#include <string>
#include <sstream>

using std::string;
using std::logic_error;
using std::ostringstream;

//debugging
#include <iostream>
using std::cout;
using std::endl;

static string makeName(Node const *node, Range const &subset_range)
{
  /* FIXME: Should be able to do this without calling C print functions 
     or using an explicit buffer 
  */
  ostringstream os;
  os << node->name() << "[";

  int ndim = subset_range.ndim(false);
  for (int i = 0; i < ndim; i++) {
    if (i > 0)
      os << ", ";
    long l = subset_range.lower()[i];
    long u = subset_range.upper()[i];
    if (l == u) {
      os << l;
    }
    else {
      os << l << ":" << u;
    }
  }
  os << "]";

  return os.str();
}

SubSetNode::SubSetNode(Node *parent, Range const &subset_range)
  : DeterministicNode(subset_range.dim(false))
{
  Range const &parent_range = parent->data.range();
  if (!parent_range.contains(subset_range)) {
    throw logic_error("Invalid subset range in SubSetNode constructor");
  }

  /* 
     This might look a little opaque, so needs explaining The values
     of a subset node can be stored in a 2 dimensional array of size
     _n X _m.  We don't need to assign storage for this array since
     the storage is already provided by the parent node. All we need
     to do is assign a vector of pointers of size _n, and this will
     behave like the array we need.

     Some examples: 

     Parent dimension [5,4,3]. Subset range [1:5, 1:4, 2]
     In this case _n = 1, _m = 5 x 4 = 20.
     _value_pointers[0] is a pointer to the first element of 
     parent->data.value(). All the other values of the subset
     are contiguous.
     
     Parent dimension [5,4,3]. Subset range [2:3, 1:4, 1:3]
     In this case _n = 4 x 3 = 12, _m = 2.
  */
  Index const &dim1 = parent_range.dim(false);
  Index const &dim2 = subset_range.dim(false);
  long ndim = parent_range.ndim(false);

  _m = 1;
  _n = dim2.length();
  for (long i = 0; i < ndim; ++i) {
    _m *= dim2[i];
    _n /= dim2[i];
    if (dim1[i] != dim2[i]) {
      break;
    }
  }
  double const ** _value_pointers = new double const *[_n];

  LeftRangeIterator l(subset_range);
  for (long i = 0; i < _n; ++i) {
    _value_pointers[i] = &parent->data.value()[parent_range.leftOffset(l)];
    for (long j = 0; j < _m; ++j) {
      ++l;
    }
  }

  string name = makeName(parent, subset_range);
  setName(name);
  addParent(parent);
}

SubSetNode::~SubSetNode()
{
  delete _value_pointers;
}

void SubSetNode::forwardSample()
{
  long length = _n * _m;
  double *value = new double[length];
  long k = 0;
  for (long i = 0; i < _n; ++i) {
    for (long j = 0; j < _m; ++j) {
      value[k++] = _value_pointers[i][j];
    }
  }
  data.setValue(value, length);
  delete [] value;
}

SubSetNode *asSubSet(Node *node)
{
  return dynamic_cast<SubSetNode*>(node);
}

bool isSubSet(Node const *node)
{
  return dynamic_cast<SubSetNode const*>(node);
}

