#include <config.h>
#include <model/ParseTree.h>
#include <model/SymTab.h>
#include <model/Compiler.h>
#include <graph/SubSetIndex.h>
#include <graph/LogicalNode.h>
#include <graph/ConstantNode.h>
#include <graph/StochasticNode.h>
#include <graph/SubSetNode.h>
#include <graph/OffsetNode.h>
#include <graph/MixtureNode.h>
#include <graph/Graph.h>
#include <distributions/DistTab.h>
#include <functions/FuncTab.h>
#include <sarray/LeftRangeIterator.h>
#include <sarray/RightRangeIterator.h>

#include <utility>
#include <vector>
#include <stdexcept>
#include <cfloat>
#include <cmath>
#include <sstream>
#include <list>

//debugging
#include <iostream>
using std::cout;
using std::endl;

#include <string>
using std::string;
using std::vector;
using std::list;
using std::map;
using std::pair;
using std::invalid_argument;
using std::runtime_error;
using std::logic_error;
using std::ostringstream;

static string cat(string const &string1, string const &string2)
{
  /* Concatenate two strings */
  string out(string1);
  out.append(string2);
  return out;
}

static long asInteger(double fval)
{
  if (fval > LONG_MAX || fval < LONG_MIN) {
    throw runtime_error("double value out of range for conversion to long");
  }
  long ival = (long) (fval + DBL_EPSILON);
  if (fabs(fval - ival) > DBL_EPSILON) {
    throw runtime_error("Invalid integer conversion");
  }
  return ival;
}

bool Compiler::constantExpression(ParseTree const *p, double &value)
{
  /* 
     Try to evaluate a constant expression.  An expression can
     only be evaluated if it is a function of
     1) Counters, or
     2) Nodes for which a data value is given.
     Functions are not allowed, but inline operators ("*", "/", "+", "-") 
     are.

     If the expression can be evaluated, then we return true and the
     expression value is written to the value argument. If it cannot
     be evaluated (because it depends on missing data) then we return
     false.
  */
  Node *node;
  Counter *counter;
  double arg1, arg2;
  std::vector<ParseTree*> const &parameters = p->parameters();

  switch (p->treeClass()) {
  case P_VAR: 
    // Is it a counter?
    counter = _countertab.getCounter(p->name());
    if (counter) {
      value = counter->value();
      return true;
    }
    else {
      // Try reading the values from the data table.
      map<string,SArray>::const_iterator i = _data_table.find(p->name());
      if (i == _data_table.end()) {
        return false;
      }
      /*
	We can't call getConstantRange() because we aren't sure that
	the range expression can be evaluated.
      */
      Range range;
      if (getRange(parameters, i->second.range(), range)) {
	// Range expression successfully evaluated
	if (range.length() > 1) {
	  throw runtime_error("Vector value in constant expression");
	}
	long offset = i->second.range().leftOffset(range.lower());
	value = i->second.value()[offset];
	if (value == JAGS_NA) {
           return false;
        } else {
	   return true;
        }
      }
      else {
	// Can't evaluate range expression
	return false;
      }
    }
    //return false;
    break;
  case P_VALUE:
    value =  p->value();
    return true;
    break;
  case P_OPERATOR:
    switch(p->getOperator()) {
    case OP_ADD:
      if (parameters.size() < 2) {
	throw logic_error("ADD must have at least two arguments in constant expression");
      }
      arg1 = 0;
      for (int i = 0; i < parameters.size(); ++i) {
	if (!constantExpression(parameters[i], arg2)) {
	  return false;
	}
	arg1 += arg2;
      }
      value = arg1;
      return true;
      break;
    case OP_SUBTRACT:
      if (parameters.size() != 2) {
	throw logic_error("SUBTRACT must have two arguments in constant expression");
      }
      if (!constantExpression(parameters[0], arg1)) {
	return false;
      }
      if (!constantExpression(parameters[1], arg2)) {
	return false;
      }
      value = arg1 - arg2;
      return true;
      break;
    case OP_MULTIPLY:
      if (parameters.size() < 2) {
	throw logic_error("MULTIPLY must have at least two arguments in constant expression");
      }
      arg1 = 1;
      for (int i = 0; i < parameters.size(); ++i) {
	if (!constantExpression(parameters[i], arg2)) {
	  return false;
	}
	arg1 *= arg2;
      }
      value = arg1;
      return true;
      break;
    case OP_DIVIDE:
      if (parameters.size() != 2) {
	throw logic_error("DIVIDE must have two arguments in constant expression");
      }
      if (!constantExpression(parameters[0], arg1)) {
	return false;
      }
      if (!constantExpression(parameters[1], arg2)) {
	return false;
      }
      value = arg1/arg2;
      return true;
      break;
    case OP_NEG:
      if (parameters.size() != 1) {
	throw logic_error("NEG must have 1 argument in constant expression");
      }
      if (!constantExpression(parameters[0], arg1)) {
	return false;
      }
      value = -arg1;
      return true;
      break;
    case OP_NONE:
      throw logic_error("Bad constant expression");
      break;
    }
    break;
  default:
    throw logic_error("Expected variable, value or expression");
  }
}

bool Compiler::indexExpression(ParseTree const *p, long &value)
{
  /* 
     Evaluate an index expression.
     
     Index expressions occur in three contexts:
     1) In the limits of a "for" loop
     2) On the left hand side of a relation
     3) On the right hand side of a relation
     
     In cases 1) and 2) we expect to be able to evaluate the index
     expression, and an exception will be thrown if it cannot be
     evaluated (strict == true). In case 3) it is not an error if an
     index expression cannot be evaluated (strict == false), but we try
     anyway for efficiency reasons.
  */

  double fvalue;
  if (constantExpression(p, fvalue) == false) {
    return false;
  }
  else {
    value = asInteger(fvalue);
    return true;
  }
}

bool Compiler::getRange(vector<ParseTree*> const &range_list, 
			Range const &default_range,
		        Range &result)
{
  /* 
     Evaluate a range expression. Returns true if successful and
     modifies the result argument.  If unsuccessful (due to missing
     values) returns false and the result argument is unchanged.
     
     The default_range argument provides default values if the range
     expression is blank: e.g. foo[] or bar[,1]
  */
  
  if (range_list.empty()) {
    // A null range expression implies the whole of the default range
    result = default_range;
    return true;
  }

  // Check size and integrity of range expression
  unsigned int size = range_list.size();
  if (size != default_range.ndim(false)) {
    throw logic_error("Default range does not match dimension of range expression");
  }
  for (int i = 0; i < size; ++i) {
    if (range_list[i]->treeClass() != P_RANGE) {
      throw logic_error("Malformed parse tree. Expected range expression");
    }
  }
  
  // Now step through and evaluate lower and upper index expressions
  Index lower(size), upper(size);
  for (unsigned int i = 0; i < size; i++) {
    switch (range_list[i]->parameters().size()) {
    case 0:
      // Empty index implies default range
      lower[i] = default_range.lower()[i];
      upper[i] = default_range.upper()[i];
      break;
    case 1:
      // Single index implies lower == upper
      if (!indexExpression(range_list[i]->parameters()[0], lower[i])) {
	return false;
      }
      else {
	upper[i] = lower[i];
      }
      break;
    case 2:
      if (!indexExpression(range_list[i]->parameters()[0], lower[i]) ||
	  !indexExpression(range_list[i]->parameters()[1], upper[i])) {
	return false;
      }
      break;
    default:
      throw logic_error("Malformed parse tree in index expression");
    }
  }
  
  // Success
  result = Range(lower, upper);
  return true;
}

Range Compiler::VariableSubsetRange(ParseTree const *var)
{
  /*
    Get the range of a subset expression of a variable on the LHS of a
    relation.  This means that the subset expression must be constant.
  */
  if (var->treeClass() != P_VAR) {
    throw logic_error("Expecting variable expression");
  }  

  string const &name = var->name();
  if (_countertab.getCounter(name) != 0) {
    // Counters are always scalar
    if (!var->parameters().empty()) {
      throw runtime_error(cat("Cannot take subset of counter ", name));
    }
    return Range();
  }
  NodeArray *array = _symtab.getVariable(name);
  if (array != 0) {
    // It's a declared node
    Range range;
    vector<ParseTree*> const &range_list = var->parameters();
    if (range_list.empty()) {
      // A null list implies the whole array
      return array->range();
    }
    else {
      if (range_list.size() != array->range().ndim(false)) {
	throw runtime_error(cat("Dimension mismatch in subset expression of array ", name));
      }
      if (!getRange(range_list, array->range(), range)) {
	throw runtime_error(cat("Missing values in subset expression of array ", name));
      }
      return range;
    }
  }
  else {
    throw runtime_error(cat("Unable to evaluate range expression: unknown variable ", name));
  }
}

Range Compiler::CounterRange(ParseTree const *var)
{
  /* The range expression for a counter differs from that of
     a variable in that it is
     1) one-dimensional
     2) may not be empty
     Further, no variables are created for counters in the
     Symbol Table
  */
  if (var->treeClass() != P_COUNTER) {
    throw logic_error("Expecting counter expression");
  }
  if (var->parameters().size() != 1) {
    throw logic_error("Invalid counter expression");
  }
  Range range();
  
  ParseTree const *prange = var->parameters()[0];
  if (prange->treeClass() != P_RANGE) {
    throw logic_error("Expecting range expression");
  }

  unsigned int size = prange->parameters().size();
  if (size < 1 || size > 2) {
    throw logic_error(cat("Invalid range expression for counter", var->name()));
  }
  long lower;
  if(!indexExpression(prange->parameters()[0], lower)) {
    throw runtime_error(cat("Unable to evaluate lower index of counter ", var->name()));
  }
  long upper;
  if (prange->parameters().size() == 2) {
    if (!indexExpression(prange->parameters()[1], upper)) {
      throw runtime_error(cat("Unable to evaluate upper index of counter ", var->name()));
    }
  }
  else {
    upper = lower;
  }
  
  /* Reverse counters that are given backwards.  NB This behaviour is
     different from BUGS, which ignores backward counters.
  */
  if (lower > upper) {
    long temp = lower;
    lower = upper;
    upper = temp;
  }

  Index ind_lower(1), ind_upper(1);
  ind_lower[0] = lower;
  ind_upper[0] = upper;

  return Range(ind_lower, ind_upper);
}

Node* Compiler::VarGetNode(ParseTree const *var)
{
  if (var->treeClass() != P_VAR) {
    throw logic_error("Expecting variable expression");
  }
  NodeArray *array = _symtab.getVariable(var->name());
  Range range = VariableSubsetRange(var);
  Node *node = array->get(range);
  
  return node;
}

void Compiler::makeSubsetIndex(ParseTree const *p, vector<SubSetIndex> &index,
			       Index const &default_index)
{
  SubSetIndex ssi;
  long i;

  if (index.size() >= default_index.length()) {
    throw logic_error("Index already full");
  }

  if (p == 0) {
    ssi = SubSetIndex(default_index[index.size()]);
  }
  else {
    switch (p->treeClass()) {
    case P_VALUE: case P_VAR: case P_OPERATOR:
      if(indexExpression(p, i)) {
	ssi = SubSetIndex(i);
      }
      else {
	ssi = SubSetIndex(getLogicalParameter(p));
      }
      break;
    default:
      throw logic_error("Expected value, variable or index expression");
    }
  }
  index.push_back(ssi);
}

Node * Compiler::getSubSetNode(ParseTree const *var)
{
  if (var->treeClass() != P_VAR) {
    throw logic_error("Expecting variable expression");
  }

  NodeArray *array = _symtab.getVariable(var->name());
  if (array == 0) {
    string msg("Unknown parameter: ");
    msg.append(var->name());
    throw runtime_error(msg);
  }
  Range subset_range;
  if (!getRange(var->parameters(), array->range(), subset_range)) {
    return 0;
  }
  return array->getSubset(subset_range);
}

Node * Compiler::getMixtureNode(ParseTree const *var)
{
  if (var->treeClass() != P_VAR) {
    throw logic_error("Expecting variable expression");
  }

  NodeArray *array = _symtab.getVariable(var->name());
  if (array == 0) {
    throw runtime_error(cat("Unknown parameter: ", var->name()));
  }
  Range const &default_range = array->range();

  vector<SubSetIndex> lower;
  vector<SubSetIndex> upper;
  vector<ParseTree*> const &range_list = var->parameters();
  int i = 0;
  for (; i < range_list.size(); ++i) {
    ParseTree const *range_element = range_list[i];
    if (range_element->treeClass() != P_RANGE) {
      throw runtime_error("Malformed range expression");
    }
    switch(range_element->parameters().size()) {
    case 0:
      /* Index is empty, implying the whole range */
      lower.push_back(default_range.lower()[i]);
      upper.push_back(default_range.upper()[i]);
      break;
    case 1:
      /* Single index: upper = lower */
      makeSubsetIndex(range_element->parameters()[0], lower, default_range.lower());
      upper.push_back(lower[i]);
      break;
    case 2:
      makeSubsetIndex(range_element->parameters()[0], lower, default_range.lower());
      makeSubsetIndex(range_element->parameters()[1], lower, default_range.lower());
      break;
    default:
      throw logic_error("Invalid range expression");
    }
  }
  if (!SubSetIndex::checkLimits(lower, upper)) {
    throw runtime_error(cat("Invalid subset expression for array ", var->name()));
  }

  //Count number of variables indices (nvi)
  int ndim = i;
  int nvi = 0;
  for (i = 0; i < ndim; ++i) {
    if (lower[i].node())
      nvi++;
  }
  if (nvi == 0) {
    throw logic_error("Trivial mixture node");
  }

  Index voffset(nvi), vlower(nvi), vupper(nvi);
  Index jl(ndim), ju(ndim);
  int j = 0;
  for (i = 0; i < ndim; ++i) {
    if (lower[i].node()) {
      voffset[j] = i;
      vlower[j] = default_range.lower()[i];
      vupper[j] = default_range.upper()[i];
      ++j;
    }
    else {
      jl[i] = lower[i].value();
      ju[i] = upper[i].value();
    }
  }

  vector<pair<long, Node*> > param;

  for (LeftRangeIterator i(Range(vlower, vupper)); !i.atEnd(); ++i) {
    for (int k = 0; k < nvi; ++k) {
      jl[voffset[k]] = i[k];
      ju[voffset[k]] = i[k];
    }
    Node *node = array->getSubset(Range(jl,ju));
    long index = array->range().leftOffset(jl);
    param.push_back(pair<long,Node*>(index, node));
  }
  
  Node *offset = new OffsetNode(default_range, lower);
  _graph.add(offset);
  
  Node *mix = new MixtureNode(offset, param);
  _graph.add(mix);

  return mix;
}

Node* Compiler::VarGetParameter(ParseTree const *var)
{
  if (var->treeClass() != P_VAR) {
    throw logic_error("Expecting variable expression");
  }
  NodeArray *array = _symtab.getVariable(var->name());
  if (array == 0) {
    return 0;
  }

  if (var->parameters().empty()) {
    // Empty range implies the whole array
    return array->getSubset(array->range());
  }
  else {
    Node *node = getSubSetNode(var);
    if (node)
      return node; //A fixed subset
    else
      return getMixtureNode(var); //A stochastic subset
  }
}

static void setConstantNameToValue(ConstantNode *node)
{
  /* Does what it says: sets the name of a constant node to
     its value (suitably formatted)
  */

  ostringstream os;
  double value = *(node->data.value());
  long ivalue = static_cast<long>(value + 1.0E-6);
  if (fabs(value - ivalue) < 1.0E-6) {
    // Integer value 
    os << ivalue;
  }
  else {
    // Floating point value
    os << value;
  }
  node->setName(os.str());
}

Node *Compiler::Param(ParseTree const *p)
{
  Node *node = 0;
  ConstantNode *cnode = 0;

  switch(p->treeClass()) {
  case P_VALUE:
    cnode = new ConstantNode(p->value());    
    setConstantNameToValue(cnode);
    node = cnode;
    _graph.add(node);
    break;
  case P_VAR:
    node = VarGetParameter(p);
    if (node == 0) {
      Counter *counter = _countertab.getCounter(p->name());
      if (counter == 0) {
	string msg("Unknown parameter: ");
	msg.append(p->name());
	throw runtime_error(msg);
      }
      node = new ConstantNode(counter->value());
      node->setName(p->name());
      _graph.add(node);
    }
    break;
  default:
    throw logic_error("Expecting value or variable expression");
  }
  return node;
}

static Function const *getLink(ParseTree const *t, FuncTab const &functab)
{
  if (t->treeClass() != P_LINK) {
    throw logic_error("Malformed parse tree: Expected link function");
  }

  Function const *func = functab.findInverse(t->name());
  if (func == 0) {
    string msg("Unable to find inverse of link function ");
    msg.append(t->name());
    throw runtime_error(msg);
  }
  else {
    return func;
  }
}

static Function const *getFunction(ParseTree const *t, FuncTab const &functab)
{
  Function const *func = 0;
  
  switch (t->treeClass()) {
  case P_FUNCTION:
    func = functab.find(t->name());
    if (func == 0) {
      string msg("Unable to find function ");
      msg.append(t->name());
      throw runtime_error(msg);
    }    
    break;
  case P_OPERATOR:
    switch(t->getOperator()) {
    case OP_ADD:
      func = functab.find("+");
      break;
    case OP_SUBTRACT:
      func = functab.find("-");
      break;
    case OP_MULTIPLY:
      func = functab.find("*");
      break;
    case OP_DIVIDE:
      func = functab.find("/");
      break;
    case OP_NEG:
      func = functab.find("UnaryMinus");
      break;
    case OP_NONE:
      throw logic_error("Bad operator expression");
      break;
    }
    if (func == 0) {
      throw logic_error("Unable to find operator");
    }
    break;
  default:
    throw logic_error("Malformed parse tree: Expected expression");
  }
  return func;
}

Node* Compiler::getLogicalParameter(ParseTree const *t)
{
  Node *node = 0;
  LogicalNode *lnode = 0;
  ConstantNode *cnode = 0;
  vector<Node*> parents;

  switch (t->treeClass()) {
  case P_VALUE:
    cnode = new ConstantNode(t->value());
    setConstantNameToValue(cnode);
    _graph.add(cnode);
    return cnode;
    break;
  case P_VAR:
    node = Param(t);
    return node;
    break;
  case P_FUNCTION:
    getLogicalParameterVector(t, parents);
    lnode = _logicalfactory.getLogicalNode(getFunction(t, _functab), parents);
    return lnode;
    break;
  case P_OPERATOR:
    getLogicalParameterVector(t, parents);
    lnode = _logicalfactory.getLogicalNode(getFunction(t,_functab), parents);
    return lnode;
    break;
  default:
    throw  logic_error("Malformed parse tree. Expected value, variable or expression");
  }
  return 0; //Wall
}

void Compiler::getLogicalParameterVector(ParseTree const *t,
					 vector<Node*> &parents)
{
  if (!parents.empty()) {
    throw logic_error("parent vector must be empty in getLogicalParameterVector");
  }

  switch (t->treeClass()) {
  case P_FUNCTION: case P_LINK: case P_OPERATOR:
    for (int i = 0; i < t->parameters().size(); ++i) {
      parents.push_back(getLogicalParameter(t->parameters()[i]));
    }
    break;
  default:
    throw logic_error("Invalid Parse Tree.  Expected function or operator.");
  }
}

void Compiler::writeLogical(ParseTree const *dtrmrel)
{
  if (dtrmrel->treeClass() != P_DETRMREL) {
    throw logic_error("Malformed parse tree. Expecting deterministic relation");
  }

  ParseTree const *var = dtrmrel->parameters()[0];
  ParseTree const *expression = dtrmrel->parameters()[1];

  Node *node = VarGetNode(var);
  LogicalNode *lnode = asLogical(node);
  if (!lnode) {
    throw logic_error(cat("Logical node undefined: ", var->name()));
  }

  vector<Node*> parameters;
  Function const *func;

  switch (expression->treeClass()) {
  case P_VALUE: case P_VAR:
    func = _functab.find("identity");
    lnode->setFunction(func);
    parameters.push_back(getLogicalParameter(expression));
    lnode->setParameters(parameters);
    break;
  case P_FUNCTION: case P_OPERATOR:
    func = getFunction(expression,_functab);
    getLogicalParameterVector(expression, parameters);
    lnode->setFunction(func);
    lnode->setParameters(parameters);
    break;
  case P_LINK:
    func = getLink(expression ,_functab);
    getLogicalParameterVector(expression, parameters);
    lnode->setFunction(func);
    lnode->setParameters(parameters);
    break;
  default:
    throw logic_error("Malformed parse tree: Expected value, variable or expression");
  }
}

void Compiler::writeStochastic(ParseTree const *stoch_relation)
{
  if (stoch_relation->treeClass() != P_STOCHREL) {
    throw logic_error("Malformed parse tree. Expecting stochastic relation");
  }

  // Get stochastic node
  ParseTree *var = stoch_relation->parameters()[0];
  Node *node = VarGetNode(var);
  StochasticNode *snode = asStochastic(node);
  if (!snode) {
    throw logic_error(cat("Stochastic node node defined: ", var->name()));
  }

  // Get the distribution
  ParseTree *distribution = stoch_relation->parameters()[1];
  if (distribution->treeClass() != P_DENSITY) {
    throw logic_error("Malformed parse tree. Expected density expression");
  }
  Distribution const *dist = _disttab.find(distribution->name());
  if (dist == 0) {
    throw runtime_error(cat("Unknown distribution: ", distribution->name()));
  }
  snode->setDistribution(dist);

  // Set the parameters of the stochastic node
  vector<ParseTree*> const &param_list = distribution->parameters();
  vector<Node*> parameters;
  for (vector<ParseTree*>::const_iterator p = param_list.begin(); p != param_list.end(); ++p) {
    parameters.push_back(Param(*p));
  }
  snode->setParameters(parameters);

  // Give upper and lower bounds, if truncated
  if (stoch_relation->parameters().size() == 3) {
    ParseTree const *truncated = stoch_relation->parameters()[2];
    Node *lBound = 0, *uBound = 0;
    ParseTree const *ll = truncated->parameters()[0];
    ParseTree const *ul = truncated->parameters()[1];
    if (ll) {
      lBound = Param(ll);
    }
    if (ul) {
      uBound = Param(ul);
    }
    snode->setBounds(lBound, uBound);
  }

}

void Compiler::writeRelations(ParseTree const *relations)
{
  Counter *counter;
  ParseTree *var;

  if (relations->treeClass() != P_RELATIONS) {
    throw logic_error("Expected relations");
  }
  vector<ParseTree*> const &relation_list = relations->parameters();
  for (vector<ParseTree*>::const_iterator p = relation_list.begin(); 
       p != relation_list.end(); ++p) 
    {
      switch ((*p)->treeClass()) {
      case P_FOR:
	var = (*p)->parameters()[0];
	relations = (*p)->parameters()[1];
	counter = _countertab.pushCounter(var->name(), CounterRange(var));
	for (counter->reset(); !counter->atEnd(); counter->increment()) {
	  writeRelations(relations);
	}
	_countertab.popCounter();
	break;
      case P_STOCHREL:
	writeStochastic(*p);
	break;
      case P_DETRMREL:
	writeLogical(*p);
	break;
      default:
	throw logic_error("Expected for loop or relation");
      }
    }
} 

void Compiler::allocateNodes(ParseTree const *relations)
{
  /* Create stochastic and logical nodes
     NB I am ruling out the data transformation trick here.
  */

  Counter *counter;
  NodeArray *array;
  Range range;
  ParseTree *var;
  Node *node;

  vector<ParseTree*> const &relation_list = relations->parameters();
  for (vector<ParseTree*>::const_iterator p = relation_list.begin(); 
       p != relation_list.end(); ++p) 
    {  
      switch ((*p)->treeClass()) {
      case P_FOR:
	var = (*p)->parameters()[0];
	relations = (*p)->parameters()[1];
	counter = _countertab.pushCounter(var->name(), CounterRange(var));
	for (counter->reset(); !counter->atEnd(); counter->increment()) {
	  allocateNodes(relations);
	}
	_countertab.popCounter();
	break;
      case P_STOCHREL:
	var = (*p)->parameters()[0];
	array = _symtab.getVariable(var->name());
	range = VariableSubsetRange(var);
	node = new StochasticNode(range.dim(false));
	array->set(node, range);
	_graph.add(node);
	break;
      case P_DETRMREL:
	var = (*p)->parameters()[0];
	array = _symtab.getVariable(var->name());
	range = VariableSubsetRange(var);
	node = new LogicalNode(range.dim(false));
	array->set(node, range);
	_graph.add(node);
	break;
      default:
	throw logic_error("Malformed parse tree. Expected a relation or for loop");
	break;
      }
    }
    }

void Compiler::fillConstants()
{
  /* 
     Any nodes which do not appear on the LHS of a relation, 
     and for which a data value is given, must be constant nodes

     NB We create the constant nodes, but don't set their values.
  */
  for (map<string, SArray>::const_iterator p(_data_table.begin());
       p != _data_table.end(); ++p) 
    {
      NodeArray *array = _symtab.getVariable(p->first);
      if (array) {
	// Skip names in the data table that are not declared
	for (LeftRangeIterator i(array->range()); !i.atEnd(); ++i) {
	  Range target_range = Range(i, i);
	  if (!array->rangeContainsSetNodes(target_range)) {
	    long j = array->range().leftOffset(i);
	    if (p->second.value()[j] != JAGS_NA) {
	      ConstantNode *cnode = new ConstantNode();
	      array->set(cnode, Range(i,i));
	      _graph.add(cnode);
	    }
	  }
	}
      }
    }
}

Compiler::Compiler(Graph &graph, SymTab &symtab,
		   FuncTab const &functab, DistTab const &disttab,
		   map<string, SArray> const &data_table)
  : _graph(graph), _symtab(symtab), _functab(functab), _disttab(disttab),
    _data_table(data_table)
{
  if (graph.size() != 0)
    throw invalid_argument("Non empty graph in Compiler constructor");

  /*FIXME: There isn't currently a way of testing the size of a SymTab
    because it contains both variables and counters. These should
    be in separate objects because the variables are persistent but
    the counters aren't */
}

void Compiler::declareVariables(vector<ParseTree*> const &dec_list)
{
  vector<ParseTree*>::const_iterator p;
  for (p = dec_list.begin() ; p != dec_list.end(); ++p) {
    if ((*p)->treeClass() != P_VAR) {
      throw invalid_argument("Expected variable expression");
    }
  }
  
  for (p = dec_list.begin() ; p != dec_list.end(); ++p) {
    ParseTree const *node_dec = *p;
    string const &name = node_dec->name();
    unsigned int ndim = node_dec->parameters().size();
    if (ndim == 0) {
      // Variable is scalar
      _symtab.addVariable(name);
    }
    else {
      // Variable is an array
      Index dim(ndim);
      for (int i = 0; i < ndim; ++i) {
	if (!indexExpression(node_dec->parameters()[i], dim[i])) {
	  throw runtime_error(cat("Unable to calculate dimensions of node ", name));
	}
      }
      _symtab.addVariable(name, dim);
    }
  }
}
  
void Compiler::writeValuesToGraph(map<std::string, SArray> const &data_table,
				    bool isdata)
    {
      for(map<string, SArray>::const_iterator p(data_table.begin()); 
	  p != data_table.end(); ++p) {
	NodeArray *array = _symtab.getVariable(p->first);
	if (array) {
      if (array->range().dim(false) != p->second.dim(false)) {
	string msg("Dimension mismatch when setting value of ");
	msg.append(p->first);
	throw runtime_error(msg);
      }
      array->setValue(p->second, isdata);
    }
  }
}


void Compiler::collectNodes()
{
  vector<Node*> nodes;
  _symtab.getNodes(nodes);
  _logicalfactory.graph().getNodes(nodes);

  for (vector<Node*>::iterator p = nodes.begin(); p != nodes.end(); ++p) {
    _graph.add(*p);
  }
}
