#ifndef LOGICAL_FACTORY_H_
#define LOGICAL_FACTORY_H_

#include <graph/Graph.h>

#include <vector>
#include <map>

class Function;
class LogicalNode;

/** 
 *  Comparison operator for a "logical pair" consisting of a function
 *  and a vector of Nodes correspoding to the parameters of the
 *  function.
 *  
 *  The pair is ordered first on the function and second on the
 *  vector of parameters, which we *assume* are the same length.
 */
struct ltlogical
{
  bool operator()(std::pair<Function const*,std::vector<Node*> > arg1,
		  std::pair<Function const*,std::vector<Node*> > arg2)
  {
    if (arg1.first < arg2.first) {
      return true;
    }
    else if (arg1.first > arg2.first) {
      return false;
    }
    else {
      //Same function. Check parameters
      unsigned long n = arg2.second.size();
      for (unsigned long i = 0; i < n; ++i) {
	if (arg1.second[i] < arg2.second[i]) {
	  return true;
	}
	else if (arg1.second[i] > arg2.second[i]) {
	  return false;
	}
      }
      return false; //All parameters equal
    }
  }
};

/**
 * @short Factory object for logical nodes 
 * The value of a logical node is determined uniquely by its function
 * and its parameters.  The purpose of a LogicalFactory is to avoid
 * unnecessary duplication of logical nodes by having a container
 * class and factory object for them that will create and/or lookup
 * logical nodes based on the function and the parameter vector
 */
class LogicalFactory { 
  Graph _graph;
  std::map<std::pair<Function const *, std::vector<Node*> >, LogicalNode*,
	   ltlogical> _logicalmap;
	
public:
  /**
   * Get a logical node with a given function and given parameters.
   * The results are cached, so if a request is repeated, the same
   * node will be returned.
   */
  LogicalNode *getLogicalNode(Function const *func, 
			      std::vector<Node*> const &param);
  /**
   * Get the graph of nodes created by the LogicalFactory
   */
  Graph const &graph() const;
};

#endif /* LOGICAL_FACTORY_H_ */
