#include <config.h>
#include <model/Model.h>
#include <model/TraceMonitor.h>
#include <sampler/ConjugateFactory.h>
#include <sampler/SliceSampler.h>
#include <sampler/FiniteSampler.h>
#include <graph/GraphMarks.h>
#include <graph/StochasticNode.h>
#include <graph/NodeError.h>
#include <graph/Node.h>
#include <graph/DeterministicNode.h>
#include <sarray/RightRangeIterator.h>
#include <sarray/LeftRangeIterator.h>

#include <Rmath.h>

#include <iostream>
#include <fstream>
#include <sstream>
#include <set>
#include <stdexcept>
#include <string>

using std::vector;
using std::list;
using std::set;
using std::ofstream;
using std::endl;
using std::cout; //debugging
using std::cerr;
using std::logic_error;
using std::runtime_error;
using std::string;
using std::ostringstream;

Model::Model()
{
  _iteration = 0;
  _initialized = false;
  _cansample = false;
}

Model::~Model()
{
  while(!_samplers.empty()) {
    Sampler *sampler = _samplers.back();
    delete sampler;
    _samplers.pop_back();
  }

  while(!_monitors.empty()) {
    TraceMonitor *monitor = _monitors.back();
    delete monitor;
    _monitors.pop_back();
  }
}

Graph &Model::graph()
{
  return _graph;
}

bool Model::isInitialized()
{
  return _initialized;
}

void Model::initialize()
{
  if (_initialized)
    throw logic_error("Model already initialized");
  if (!_cansample)
    throw logic_error("Graph not checked yet");

  /* Count the uninitialized nodes */
  unsigned int n = _nodes.size();
  for (vector<Node*>::iterator i = _nodes.begin(); i != _nodes.end(); i++)
    {
      if ((*i)->isInitialized())
	n--;
    }

  for (vector<Node*>::iterator i = _nodes.begin(); i != _nodes.end(); i++)
    {
      Node *node = *i;
      if (!node->isInitialized()) {
	node->initialize();
	if (!node->isInitialized()) {
	  throw NodeError(node, "Initialization failure");
	}
	n--;
      }
    }

  if (n != 0) {
    throw logic_error("Model::initialize failed");
  }

  /* Now find fixed deterministic nodes 
     THIS IS REDUNDANT: Already taken care of by DeterministicNode::initialize
  for (vector<Node*>::iterator i = _nodes.begin(); i != _nodes.end(); 
       i++)
    {
      if (isDeterministic(*i) && !isObserved(*i)) {	
	// A deterministic node is fixed if all of its parents are fixed.
	set<Node*> const &parents = (*i)->parents();
	bool canfix = !parents.empty();
	for (set<Node*>::iterator j = parents.begin(); 
	     j != parents.end(); ++j) {
	  if (!(*j)->data.isFixed()) {
	    canfix = false;
	    break;
	  }
	}
	if (canfix) {
	  cout << "FIXING " << (*i)->name() << endl;
	  (*i)->data.setFixed(true);
	}
      }
    }
  */
  _initialized = true;
}

void Model::checkGraph()
{
  if (_cansample) {
    throw logic_error("Already checked ability to sample");
  }
  if (!_graph.isClosed())
    throw runtime_error("Graph not closed");
  if (_graph.hasCycle()) 
    throw runtime_error("Directed cycle in graph");
  if (_nodes.empty()) {
    _graph.getSortedNodes(_nodes);
  }

  _cansample = true;
}

void Model::chooseSamplers(vector<SamplerFactory const *> const &samplers)
{
  if (!_samplers.empty())
    throw logic_error("Samplers already chosen");
  if (!_cansample) 
    throw logic_error("Graph not checked");

  /**
   * Get the set of all informative stochastic nodes
   */
  set<StochasticNode*> stochastic_nodes;
  for (vector<Node*>::iterator i = _nodes.begin(); i != _nodes.end(); i++) {
    StochasticNode *snode = asStochastic(*i);
    if (snode && !snode->data.isFixed() && snode->hasDescendant(isObserved))
      {
	stochastic_nodes.insert(snode);
      }
  }

  /**
   * Traverse the list of samplers, selecting nodes that can be sampled
   */
  for (vector<SamplerFactory const *>::const_iterator p = samplers.begin();
       p != samplers.end(); ++p)
    {
      (*p)->makeSampler(stochastic_nodes, _samplers);
    }
  if (!stochastic_nodes.empty()) {
    throw NodeError(*stochastic_nodes.begin(),
                    "Unable to find appropriate sampler");
  }

  /* 
     Create a graph containing all non-fixed nodes in the graph.  This
     graph will eventually contain only the nodes to be updated by the
     model using forward sampling.
  */ 
  Graph egraph; 
  for (vector<Node*>::iterator i = _nodes.begin(); i != _nodes.end(); i++)
    {
      if (!(*i)->data.isFixed()) {
	egraph.add(*i);
      }
    }
  /* Now remove nodes that are updated by a sampler */
  for (list<Sampler*>::const_iterator s = _samplers.begin();
       s != _samplers.end(); ++s) {
    vector<StochasticNode*> const &snodes = (*s)->nodes();
    for (vector<StochasticNode*>::const_iterator p(snodes.begin()); 
	 p != snodes.end(); ++p) {
      egraph.remove(*p);
    }
    vector<DeterministicNode*> const &dnodes = (*s)->deterministicChildren();
    for (vector<DeterministicNode*>::const_iterator p(dnodes.begin()); 
	 p != dnodes.end(); ++p) {
      egraph.remove(*p);
    }
  }
  egraph.getSortedNodes(_extra_nodes);
}

void Model::update(long niter)
{
  if (!_initialized) {
    throw logic_error("Attempt to update uninitialized model");
  }
  list<Sampler*>::reverse_iterator i;
  list<TraceMonitor*>::reverse_iterator j;  

  for (int iter = 0; iter < niter; ++iter) {
    for (i = _samplers.rbegin(); i != _samplers.rend(); i++) {
      (*i)->update();
    }
    /* 
       Update the uninformative nodes This only needs to be done once
       per iteration
    */
    for (vector<Node*>::iterator k = _extra_nodes.begin(); 
	 k != _extra_nodes.end(); ++k) {
      (*k)->forwardSample();
    }
    _iteration++;
    for (j = _monitors.rbegin(); j != _monitors.rend(); j++) {
      (*j)->update(_iteration);
    }
  }
}

long Model::iteration() const
{
  return _iteration;
}

TraceMonitor const * Model::setMonitor(Node *node, int thin)
{
  // Check that node is not already being monitored
  for (list<TraceMonitor*>::const_iterator j = _monitors.begin();
       j != _monitors.end(); j++) 
    {
      if ((*j)->node() == node) {
	throw NodeError(node, "Node already being monitored");
      }
    }

  TraceMonitor *monitor = new TraceMonitor(node, _iteration + 1, thin);
  _monitors.push_back(monitor);
  return monitor;
}

void Model::clearMonitor(Node const *node)
{
  for (list<TraceMonitor*>::iterator j = _monitors.begin();
       j != _monitors.end(); j++) 
    {
      if ((*j)->node() == node) {
	_monitors.erase(j);
	return;
      }
    }
}

static string printIndex (string const &name, Index const &index)
{
  ostringstream ostr;
  ostr << name << "[";
  for (unsigned int i = 0; i < index.length(); ++i) {
    if (i > 0) 
      ostr << ",";
    ostr << index[i];
  }
  ostr << "]";
  return ostr.str();
}

/* FIXME: Shameless cut-and-paste from ../Console.cc */
static void writeDouble(double x, std::ostream &out)
{
  if (x == JAGS_NA) {
    out << "NA";
  }
  else if (ISNAN(x)) {
    out << "NaN";
  }
  else if (!R_FINITE(x)) {
    if (x > 0) {
      out << "Inf";
    }
    else {
      out << "-Inf";
    }
  }
  else {
    out << x;
  }
}

void Model::coda(vector<Node const*> const &nodes, std::ostream &output, 
		 std::ostream &index)
{
  /* 
     Dump monitor for given node in CODA format.

     CODA requires two files, the output file (output) and the index
     file (index)
  */
  long lineno = 0;
  for (unsigned int i = 0; i < nodes.size(); i++) {
    for (list<TraceMonitor*>::const_iterator j = _monitors.begin();
	 j != _monitors.end(); j++) 
      {
	Node const *node = nodes[i];
	if ((*j)->node() == node) {
	  TraceMonitor const *monitor = *j;
	  int nvar = node->data.length();
	  if (nvar != 1) {
	    /* Multivariate node */
	    for (LeftRangeIterator i(node->data.range()); !i.atEnd(); ++i) {
	      index << printIndex(node->name(), i) << "  " 
		    << lineno + 1 << "  " 
		    << lineno + monitor->size() << endl;
	    
	      long iter = monitor->start();
	      long offset = node->data.range().leftOffset(i);
	      for (int k = 0; k < monitor->size(); k++) {
		output << iter << "  ";
		writeDouble(monitor->values()[k * nvar + offset], output);
		output << endl;
		iter += monitor->thin();
	      }
	      lineno += monitor->size();
	    }
	  }
	  else {
	    /* Univariate node */
	    index << node->name() << "  " << lineno + 1 << "  " 
		  << lineno + monitor->size() << endl;
	    long iter = monitor->start();
	    for (int k = 0; k < monitor->size(); k++) {
	      output << iter << "  ";
	      writeDouble(monitor->values()[k], output);
	      output << endl;
	      iter += monitor->thin();
	    }
	    lineno += monitor->size();
	  }
	}
      }
  }
}

list<TraceMonitor*> const &Model::monitors() const
{
  return _monitors;
}
