#ifndef MODEL_H_
#define MODEL_H_

#include <graph/Graph.h>

#include <vector>
#include <list>

class Sampler;
class TraceMonitor;
class SamplerFactory;

/**
 * @short Graphical model 
 *
 * The purpose of the model class is to collect together all the
 * elements necessary to run a MCMC sampler on a graphical model.
 */
class Model {
  Graph _graph;
  long _iteration;
  std::vector<Node*> _nodes;
  std::vector<Node*> _extra_nodes;
  std::list<Sampler*> _samplers;
  std::list<TraceMonitor*> _monitors;
  bool _initialized;
  bool _cansample;
public:
  Model();
  ~Model();

  Graph &graph();
  /**
   * Checks that the graph is closed and acyclic. A runtime_error
   * is thrown if it is not
   */
  void checkGraph();
  /**
   * Initializes the model.  Nodes that have not had their values set
   * are initialized by a call to the Node::initialize function (which
   * typically will do forward sampling).  
   */
  void initialize();
  /** Returns true if the model has been initialized */
  bool isInitialized();
  /** Selects samplers for each stochastic node in the graph 
   * @param samplers vector of Sampler Factories. Each factory is
   * tried in turn. If there are any informative stochastic nodes
   * left without samplers after all factories have been tried,
   * then a runtime error is thrown
   */
  void chooseSamplers(std::vector<SamplerFactory const*> const &samplers);
  /** 
   * Updates the model. A logic_error is thrown if the model is
   * uninitialized.
   * @param niter Number of iterations to run
   */
  void update(long niter);
  /**
   * Returns the current iteration number
   */
  long iteration() const;
  /**
   * Creates a monitor for the given node, with given thinning interval
   * If the node is already being monitored, a NodeError is thrown.
   */
  TraceMonitor const * setMonitor(Node *node, int thin);
  /**
   * Clears the monitor associated with the given node. If there is
   * no such monitor, then the function has no effect
   */
  void clearMonitor(Node const *node);
  /**
   * Writes the monitored values for the given vector of nodes in
   * the form used by the CODA program
   * @param nodes Vector of nodes to monitor
   * @param out Output stream for the monitored values
   * @param ind Output stream for the index
   */
  void coda(std::vector<Node const *> const &nodes,
	    std::ostream &out, std::ostream &ind);
  /**
   * Returns the list of monitors
   */
  std::list<TraceMonitor*> const &monitors() const;
};

#endif /* MODEL_H_ */
