#include <config.h>
#include <model/NodeArray.h>
#include <graph/ConstantNode.h>
#include <graph/AggNode.h>
#include <sarray/LeftRangeIterator.h>
//Need these for isStochastic, isDeterministic
#include <graph/StochasticNode.h>
#include <graph/DeterministicNode.h>
#include <graph/NodeError.h>

#include <iostream>
#include <sstream>
#include <string>
#include <stdexcept>

using std::vector;
using std::map;
using std::string;
using std::cout;
using std::endl;
using std::ostringstream;
using std::runtime_error;
using std::logic_error;

NodeArray::NodeArray(Index const &dim)
  : _range(Index(dim.length()), dim)
{
  long length = _range.length();
  _node_pointers = new Node *[length];
  _offsets = new long[length];
  for (long i = 0; i < length; i++) {
    _node_pointers[i] = 0;
    _offsets[i] = -1;
  }
}

NodeArray::~NodeArray()
{
  delete [] _node_pointers;
  delete [] _offsets;
}

void NodeArray::setName(string const &name)
{
  _name = name;
}

string const &NodeArray::name() const
{
  return _name;
}

bool NodeArray::rangeContainsSetNodes(Range const &target_range) const
{
  if (!_range.contains(target_range))
    throw logic_error("Range error in NodeArray::rangeContainsSetNodes"); 

  for (LeftRangeIterator i(target_range); !i.atEnd(); ++i) {
    if (_node_pointers[_range.leftOffset(i)] != 0)
      return true;
  }
  return false;
}

/*
static string makeName(string const &var, Range const &range)
{
  char buf[100];

  string name(var);
  name.append("[");
  unsigned int ndim = range.ndim(false);
  for (unsigned int i = 0; i < ndim; i++) {
    snprintf(buf, 100,"%ld", range.lower()[i]);
    name.append(buf);
    if (range.lower()[i] < range.upper()[i]) {
      name.append(":");
      snprintf(buf, 100,"%ld", range.upper()[i]);
      name.append(buf);
    }
    if (i < ndim - 1) {
      name.append(",");
    }
  }
  name.append("]");
  return name;
}
*/

static string printRange(string const &name, Range const &range)
{
  ostringstream ostr;
  ostr << name << "[";
  for (unsigned int i = 0; i < range.ndim(false); ++i) {
    if (i > 0)
      ostr << ",";
    if (range.lower()[i] == range.upper()[i]) {
      ostr << range.lower()[i];
    }
    else {
      ostr << range.lower()[i] << ":" << range.upper()[i];
    }
  }
  ostr << "]";
  return ostr.str();
}

void NodeArray::set(Node *node, Range const &target_range)
{
  /* Set target_range to refer to a multivariate node */

  if (!node) {
    string msg("Attempt to set NULL node to ");
    msg.append(printRange(name(), target_range));
    throw logic_error(msg);
  }
  if (node->data.dim(true) != target_range.dim(true)) {
    string msg("Cannot set ");
    msg.append(printRange(name(), target_range));
    msg.append(". Dimension mismatch");
    throw runtime_error(msg);
  }
  if (!_range.contains(target_range)) {
    string msg("Cannot set ");
    msg.append(printRange(name(), target_range));
    msg.append(". Range out of bounds");
    throw runtime_error(msg);
  }
  if (rangeContainsSetNodes(target_range)) {
    string msg("Cannot set ");
    msg.append(printRange(name(), target_range));
    msg.append(". Range already contains set nodes");
    throw runtime_error(msg);
  }

  /* Add the node to the table of set nodes */
  _set_nodes[target_range] = node;

  /* Set the node name */
  if (!this->name().empty() && node->name().empty()) {
    if (target_range == this->range()) {
      node->setName(this->name());
    }
    else {
      //string node_name = makeName(this->name(), target_range);
      string node_name = printRange(this->name(), target_range);
      node->setName(node_name);
    }
  }

  /* Set the _node_pointers array and the offset array */
  Range const &node_range = node->data.range();
  LeftRangeIterator i(node_range), j(target_range);
      
  for (; !j.atEnd(); ++i, ++j)
    {
      long offset = _range.leftOffset(j);
      _node_pointers[offset] = node;
      _offsets[offset] = node_range.leftOffset(i);
    }

  /* Add to the graph */
  _graph.add(node);
}

/*
void NodeArray::set(Node *node, Index const &target_index)
{
  Range target_range(target_index, target_index);
  return set(node, target_range);
}
*/

Node *NodeArray::get(Range const &target_range) const
{
  /* Returns a previously set node. The target range must correspond exactly
     to the range in a previous call to set(). */

  if (!_range.contains(target_range)) {
    //warn("Cannot get range from array node. Range out of bounds");
    return 0;
  }

  map<Range, Node *, ltrange>::const_iterator p 
    = _set_nodes.find(target_range);

  if (p != _set_nodes.end()) {
    return p->second;
  }
  else {
    return 0;
  }
}

Node *NodeArray::getSubset(Range const &target_range)
{
  /* If range corresponds to a set node, then return this */
  Node *node = get(target_range);
  if (node)
    return node;

  /* If range corresponds to a previously created subset, then return this */
  std::map<Range, Node *, ltrange>::iterator p
    = _generated_nodes.find(target_range);
  if (p != _generated_nodes.end()) {
    return p->second;
  }

  /* Otherwise create an aggregate node */
  vector<Node *> nodes;
  vector<long> offsets;
  for (LeftRangeIterator i(target_range); !i.atEnd(); ++i) {
    long offset = _range.leftOffset(i);
    if (_node_pointers[offset] == 0) {
      /* Create a constant node to plug the gap */
      ConstantNode *cnode = new ConstantNode();
      this->set(cnode, Range(i,i));
    }
    if (_node_pointers[offset] == 0) {
      throw logic_error("Internal error in NodeArray:getSubset");
    }
    nodes.push_back(_node_pointers[offset]);
    offsets.push_back(_offsets[offset]);
  }
  node = new AggNode(target_range.dim(false), nodes, offsets);
  if (!this->name().empty()) {
    if (target_range == this->range()) {
      node->setName(this->name());
    }
    else {
      //string node_name = makeName(this->name(), target_range);
      string node_name = printRange(this->name(), target_range);
      node->setName(node_name);
    }
  }
  _generated_nodes[target_range] = node;

  _graph.add(node);
  return node;
}

void NodeArray::setValue(SArray const &value, bool observed)
{
  if (!(value.range() == _range)) {
    string msg("Dimension mismatch when setting value of node array ");
    msg.append(name());
    throw runtime_error(msg);
  }

  /* Slightly extravagant: We assign an array of maximal size
     that will be large enough to contain the values of any member node  */
  double *member_value = new double[value.length()];

  map<Range, Node *, ltrange>::const_iterator p = _set_nodes.begin();
  for (; p != _set_nodes.end(); ++p) {
    int j = 0;
    for (LeftRangeIterator i(p->first); !i.atEnd(); ++i) {
      member_value[j++] = value.value()[_range.leftOffset(i)];
    }
    long member_length = j;
    /* Check to see if value vector contains missing values */
    bool missing = (member_value[0] == JAGS_NA);
    for (j = 0; j < member_length; j++) {
      if ((member_value[j] == JAGS_NA) != missing) {
	Node *node = p->second;
	throw NodeError(p->second, "Values supplied for node are partially missing");
      }
    }
    if (!missing) {
      Node *node = p->second;
      if (isDeterministic(node)) {
         throw NodeError(p->second, "Cannot set value of deterministic node");
      }
      node->data.setValue(member_value, member_length);
      if (observed) {
      node->data.setFixed(true);
      }
    }
  }
  delete [] member_value;
}

void NodeArray::getValue(SArray &value, bool observed) const
{
  if (!(value.range() == _range)) {
    string msg("Dimension mismatch when getting value of node array ");
    msg.append(name());
    throw runtime_error(msg);
  }

  long array_length = value.length();
  double *array_value = new double[array_length];
  for (int j = 0; j < array_length; j++) {
    array_value[j] = JAGS_NA;
  }

  map<Range, Node *, ltrange>::const_iterator p = _set_nodes.begin();
  for (; p != _set_nodes.end(); ++p) {
    Node *node = p->second;
    if (observed == isObserved(node)) {
      if (isStochastic(node) || (isConstant(node) && observed == true)) {
	int j = 0;
	for (LeftRangeIterator i(p->first); !i.atEnd(); ++i) {
	  array_value[_range.leftOffset(i)] = p->second->data.value()[j++];
	}
      }
    }
  }

  value.setValue(array_value, array_length);
  delete [] array_value;
}

Range const &NodeArray::range() const
{
  return _range;
}

Graph const &NodeArray::graph() const
{
  return _graph;
}
