#include <config.h>
#include <sampler/ConjugateBeta.h>
#include <graph/LogicalNode.h>
#include <graph/StochasticNode.h>

#include <Rmath.h>

#include <set>
#include <stdexcept>
#include <vector>
#include <cmath>
#include <algorithm>

using std::vector;
using std::set;
using std::sqrt;
using std::invalid_argument;
using std::logic_error;
using std::max;
using std::min;

ConjugateBeta::ConjugateBeta(StochasticNode *snode)
  : ConjugateSampler(snode) 
{
  if (!canSample(snode)) {
    throw invalid_argument("Can't construct ConjugateBeta sampler");
  }
}

ConjugateBeta::~ConjugateBeta()
{

}

bool ConjugateBeta::canSample(StochasticNode const *snode)
{
  if (getDist(snode) != BETA)
    return false;

  vector<StochasticNode*> stoch_nodes;
  vector<DeterministicNode*> dtrm_nodes;
  classifyChildren(snode, stoch_nodes, dtrm_nodes);
  /* 
     Create a set of nodes containing snode and its deterministic
     descendants for the checks below.
  */
  set<SArray const *> paramset;
  paramset.insert(&snode->data);
  for (unsigned int j = 0; j < dtrm_nodes.size(); ++j) {
    paramset.insert(&dtrm_nodes[j]->data);
  }

  // Check stochastic children
  for (unsigned int i = 0; i < stoch_nodes.size(); ++i) {
    if (stoch_nodes[i]->isBounded()) {
      return false; //Bounded
    }
    switch(getDist(stoch_nodes[i])) {
    case BIN: 
      if (paramset.count(stoch_nodes[i]->parameters()[1])) {
	return false; //n depends on snode
      }      
      break;
    case BERN:
      break;
    default:
      return false;
    }
  }
  
  // Check deterministic descendants
  for (unsigned int j = 0; j < dtrm_nodes.size(); ++j) {
    if (dtrm_nodes[j]->hasDescendant(isStochastic)) {
      // We only allow direct children
      return false;
    }
  }

  return true; //We made it!
}

void ConjugateBeta::update()
{
  vector<StochasticNode*> const &stoch_children = stochasticChildren();
  StochasticNode const *snode = node();

  double a = *snode->parameters()[0]->value();
  double b = *snode->parameters()[1]->value();

  for (unsigned int i = 0; i < stoch_children.size(); ++i) {
    StochasticNode const *schild = stoch_children[i];
    double y = *schild->data.value();
    double n;
    switch(getDist(schild)) {
    case BIN:
      n = *schild->parameters()[1]->value();
      break;
    case BERN:
      n = 1;
      break;
    default:
      throw logic_error("Invalid distribution in Conjugate Beta sampler");
    }
    a += y;
    b += (n - y);
  }

  // Draw the sample
  double xnew = rbeta(a, b);
  if (node()->isBounded()) {
    double lower = 0;
    SArray const *lb = node()->lowerBound();
    if (lb) {
      lower = max(lower, *lb->value());
    }
    double upper = 1;
    SArray const *ub = node()->upperBound();
    if (ub) {
      upper = min(upper, *ub->value());
    }
    /* Try 4 more attempts to get random sample within the bounds */
    for (int i = 0; i < 4; i++) {
      if (xnew >= lower && xnew <= upper) {
	setValue(&xnew, 1);
	return;
      }
      xnew = rbeta(a, b);
    }
    /* Failure! Use inversion */
    double plower = lb ? pbeta(lower, a, b, 1, 0) : 0;
    double pupper = ub ? pbeta(upper, a, b, 1, 0) : 1;
    double p = runif(plower, pupper);
    xnew = qbeta(p, a, b, 1, 0);   
  }
  setValue(&xnew, 1);
}
