#include <config.h>
#include <sampler/ConjugateGamma.h>
#include <graph/LogicalNode.h>
#include <graph/StochasticNode.h>
#include <graph/MixtureNode.h>

#include <Rmath.h>

#include <set>
#include <stdexcept>
#include <vector>
#include <cmath>
#include <algorithm>

using std::vector;
using std::set;
using std::sqrt;
using std::invalid_argument;
using std::logic_error;
using std::max;

//debuggin
#include <iostream>
using std::cout;
using std::endl;

ConjugateGamma::ConjugateGamma(StochasticNode *snode)
  : ConjugateSampler(snode)
{
  if (!canSample(snode)) {
    throw invalid_argument("Can't construct ConjugateGamma sampler");
  }
  _fast = deterministicChildren().empty();
}

ConjugateGamma::~ConjugateGamma()
{
}

bool ConjugateGamma::canSample(StochasticNode const *snode)
{
  /*
    The exponential and chisquare distributions are both special
    cases of the gamma distribution and are handled by the conjugate
    gamma sampler.
  */

  switch (getDist(snode)) {
  case GAMMA: case EXP: case CHISQ:
    break;
  default:
    return false;
  }

  vector<StochasticNode*> stoch_nodes;
  vector<DeterministicNode*> dtrm_nodes;
  classifyChildren(snode, stoch_nodes, dtrm_nodes);
  /* 
     Create a set of nodes containing snode and its deterministic
     descendants for the checks below.
  */
  set<SArray const *> nodeset;
  nodeset.insert(&snode->data);
  for (unsigned int j = 0; j < dtrm_nodes.size(); ++j) {
    nodeset.insert(&dtrm_nodes[j]->data);
  }

  // Check stochastic children
  for (unsigned int i = 0; i < stoch_nodes.size(); ++i) {
    vector<SArray const *> const &param = stoch_nodes[i]->parameters();
    if (stoch_nodes[i]->isBounded()) {
      return false; //Bounded
    }
    switch(getDist(stoch_nodes[i])) {
    case EXP: case POIS:
      break;
    case GAMMA: case NORM: case DEXP: case WEIB:
      if (nodeset.count(param[0])) {
	return false; //non-scale parameter depends on snode
      }
      break;
    default:
      return false;
    }
  }
  
  // Check deterministic descendants
  for (unsigned int j = 0; j < dtrm_nodes.size(); ++j) {
    LogicalNode const *lnode = asLogical(dtrm_nodes[j]);
    if (lnode) {
      vector<SArray const *> const &param = lnode->parameters();
      switch(getOp(lnode)) {
      case IDENTITY:
	break;
      case DIVIDE:
	if (nodeset.count(param[1]))
	  return false; //reciprocal term
	break;
      case MULTIPLY:
	if (nodeset.count(param[0]) && nodeset.count(param[1]))
	  return false; //quadratic term
	break;
      default:
	return false;
      }
    }
    else if (isMixture(dtrm_nodes[j])) {
      // Check that index does not depend on snode
      MixtureNode const *mnode = asMixture(dtrm_nodes[j]);
      if (nodeset.count(&mnode->index()->data))
	return false;
    }
    else {
      return false;
    }
  }

  return true; //We made it!
}

void ConjugateGamma::update()
{
  vector<StochasticNode*> const &stoch_children = stochasticChildren();
  unsigned int nchildren = stoch_children.size();

  double r; // shape
  double mu; // 1/scale
  vector<SArray const *> const &param = node()->parameters();
  switch(getDist(node())) {
  case GAMMA:
    r = *param[0]->value();
    mu = *param[1]->value();
    break;
  case EXP:
    r = 1;
    mu = *param[0]->value();
    break;
  case CHISQ:
    r = *param[0]->value()/2;
    mu = 1/2;
    break;
  default:
    throw logic_error("invalid distribution in ConjugateGamma sampler");
  }
  if (_fast) {
    for (unsigned int i = 0; i < nchildren; ++i) {
      vector<SArray const *> const &cparam = stoch_children[i]->parameters();
      StochasticNode const *schild = stoch_children[i];
      double Y = *schild->data.value();
      double ymean; //normal mean
      switch(getDist(schild)) {
      case GAMMA:
	r += *cparam[0]->value();
	mu += Y;
	break;
      case EXP:
	r += 1;
	mu += Y;
	break;
      case NORM:
	r += 0.5;
	ymean = *cparam[0]->value();
	mu += (Y - ymean) * (Y - ymean) / 2;
	break;
      case POIS:
	r += Y;
	mu += 1;
	break;
      case DEXP:
	r += 1;
	ymean = *cparam[0]->value();
	mu += fabs(Y - ymean);
	break;
      case WEIB:
	r += 1; 
	mu += pow(Y, *cparam[0]->value());
	break;
      default:
	throw logic_error("Invalid distribution in Conjugate Gamma sampler");
      }
    }
  }
  else {
    double *C0 = new double[nchildren];
    double val = 0;    
    setValue(&val, 1);
    for (unsigned int i = 0; i < nchildren; ++i) {
      vector<SArray const *> const &cparam = stoch_children[i]->parameters();
      StochasticNode const *schild = stoch_children[i];
      switch(getDist(schild)) {
      case GAMMA: case NORM: case DEXP: case WEIB:
	C0[i] = *cparam[1]->value();
	break;
      case EXP: case POIS:
	C0[i] = *cparam[0]->value();
	break;
      default:
	throw logic_error("Invalid distribution in Conjugate Gamma sampler");
      }
    }
    val = 1;
    setValue(&val, 1);
    for (unsigned int i = 0; i < nchildren; ++i) {
      vector<SArray const *> const &cparam = stoch_children[i]->parameters();
      StochasticNode const *schild = stoch_children[i];
      double C;
      double Y = *schild->data.value();
      double ymean; //normal mean
      switch(getDist(schild)) {
      case GAMMA: case NORM: case DEXP: case WEIB:
	C = *cparam[1]->value();
	break;
      case EXP: case POIS:
	C = *cparam[0]->value();
	break;
      default:
	throw logic_error("Invalid distribution in Conjugate Gamma sampler");
      }
      //      if (fabs(C - C0[i]) >  1.0E-6) {
      if (fabs(C - C0[i]) >  DBL_EPSILON) {
	switch(getDist(schild)) {
	case GAMMA:
	  r += *cparam[0]->value();
	  mu += C * Y;
	  break;
	case EXP:
	  r += 1;
	  mu += C * Y;
	  break;
	case NORM:
	  r += 0.5;
	  ymean = *cparam[0]->value();
	  mu += C * (Y - ymean) * (Y - ymean) / 2;
	  break;
	case POIS:
	  r += Y;
	  mu += C;
	  break;
	case DEXP:
	  r += 1;
	  ymean = *cparam[0]->value();
	  mu += C * fabs(Y - ymean);
	  break;
	case WEIB:
	  r += 1; 
	  mu += C * pow(Y, *cparam[0]->value());
	  break;
	default:
	  throw logic_error("Invalid distribution in Conjugate Gamma sampler");
	}
      }
    }
  }

  double xnew;
  if (node()->isBounded()) {
    /* Use inversion to get random sample */
    double lower = 0;
    SArray const *lb = node()->lowerBound();
    if (lb) {
      lower = max(lower, *lb->value());
    }
    SArray const *ub = node()->upperBound();
    double plower = lb ? pgamma(lower,        r, 1/mu, 1, 0) : 0;
    double pupper = ub ? pgamma(*ub->value(), r, 1/mu, 1, 0) : 1;
    double p = runif(plower, pupper);
    xnew = qgamma(p, r, 1/mu, 1, 0);    
  }
  else {
    xnew = rgamma(r, 1/mu);
  }
  setValue(&xnew, 1);  
}
