#include <config.h>
#include <sampler/FiniteSampler.h>
#include <distributions/DistFinite.h>
#include <graph/StochasticNode.h>

#include <Rmath.h>

#include <cmath>
#include <string>
#include <stdexcept>

using std::logic_error;
using std::string;

FiniteSampler::FiniteSampler(StochasticNode *node)
  : GibbsSampler(node)
{
  DistFinite const *fdist = asFinite(node->distribution());
  if (fdist == 0) {
    string msg("Attempt to use finite sampler on node ");
    msg.append(node->name());
    msg.append(" with non-finite distribution");
    throw logic_error(msg);
  }
}

FiniteSampler::~FiniteSampler()
{}

void FiniteSampler::update()
{
  DistFinite const *fdist = asFinite(node()->distribution());

  long lower = fdist->lowerLimit(node()->parameters());
  long upper = fdist->upperLimit(node()->parameters());

  long size = upper - lower + 1;
  double *loglik = new double[size];

  for (long i = 0; i < size; i++) {
    double ivalue = lower + i;
    setValue(&ivalue, 1);
    loglik[i] = logFullConditional();
  }

  /* Exponentiate to get values proportional to probabilities */
  double llsum = 0.0;
  for (long i = 0; i < size; i++) {
    loglik[i] = exp(loglik[i]);
    llsum += loglik[i];
  }

  /* Sample */
  double urand = runif(0.0, llsum);
  long i;
  llsum = 0.0;
  for (i = 0; i < size - 1; i++) {
    llsum += loglik[i];
    if (llsum > urand) {
      break;
    }
  }
  double ivalue = lower + i;
  setValue(&ivalue, 1);
  delete [] loglik;
}

bool FiniteSampler::canSample(StochasticNode const *node)
{
  DistFinite const *fdist = asFinite(node->distribution());
  if (fdist == 0)
    return false;
  else
    return true;
}
