#include <config.h>
#include <sampler/Sampler.h>
#include <graph/StochasticNode.h>
#include <graph/DeterministicNode.h>
#include <graph/Graph.h>
#include <graph/NodeError.h>

#include <Rmath.h>

#include <stdexcept>
#include <set>
#include <string>

using std::vector;
using std::set;
using std::runtime_error;
using std::logic_error;
using std::string;

//debbugin
#include <iostream>

Sampler::Sampler(vector<StochasticNode *> const &nodes)
  : _nodes(nodes)
{
  set<StochasticNode const *> nodeset;
  for (vector<StochasticNode *>::const_iterator p = nodes.begin();
       p != nodes.end(); ++p) {
    nodeset.insert(*p);
  }
  classifyChildren(nodeset, _stochastic_children, _deterministic_children);
}

Sampler::~Sampler()
{
  // virtual destructor
}

vector<StochasticNode *> const &Sampler::nodes() const
{
  return _nodes;
}

static void classifyNode(Node *node, Graph &sgraph, Graph &dgraph)
{
  if (isStochastic(node)) {
    if (isObserved(node) || node->hasDescendant(isObserved)) {
      sgraph.add(node);
    }
  }
  else if (isDeterministic(node)) {
    if (node->hasDescendant(isObserved)) {
      dgraph.add(node);      
      for (set<Node*>::iterator p = node->children().begin();
	   p != node->children().end(); ++p) {
	classifyNode(*p, sgraph, dgraph);
      }
    }
  }
  else {
    string msg("Unable to classify node ");
    msg.append(node->name());
    throw logic_error(msg);
  }
}

void Sampler::classifyChildren(set<StochasticNode const *> const &nodes,
			       vector<StochasticNode*> &stoch_nodes,
			       vector<DeterministicNode*> &dtrm_nodes)
{
  Graph sgraph, dgraph;
  
  /* Classify children of each node */
  for (set<StochasticNode const *>::const_iterator p = nodes.begin();
       p != nodes.end(); ++p) 
    {
      StochasticNode const *node = *p;
      for (set<Node*>::const_iterator q = node->children().begin(); 
	   q != node->children().end(); ++q) 
	{
	  classifyNode(*q, sgraph, dgraph);
	}
    }

  vector<Node*> svector;
  sgraph.getNodes(svector);
  for (vector<Node*>::iterator i = svector.begin(); i != svector.end(); 
       ++i) 
    {
      StochasticNode *snode = asStochastic(*i);
      if (nodes.count(snode) == 0) {
	stoch_nodes.push_back(snode);
      }
    }
  
  vector<Node*> dvector;
  dgraph.getSortedNodes(dvector);
  for (vector<Node*>::iterator i = dvector.begin(); i != dvector.end(); 
       ++i) 
    {
      dtrm_nodes.push_back(asDeterministic(*i));
    }
}

double Sampler::logFullConditional()
{
  double logprior = 0;
  for (vector<StochasticNode*>::const_iterator p(_nodes.begin());
       p != _nodes.end(); ++p) 
    {
      logprior += (*p)->logDensity();
    }

  double loglikelihood = 0;
  for (vector<StochasticNode*>::const_iterator p(_stochastic_children.begin());
       p != _stochastic_children.end(); ++p) 
    {
      double l = (*p)->logDensity();
      if (l == -DBL_MAX || !R_FINITE(l)) {
	return l;
      }
      else {
	loglikelihood += l;
      }
    }

  if (ISNAN(loglikelihood)) {
    /* Go back and find the first node to fail */
    for (vector<StochasticNode*>::const_iterator p(_stochastic_children.begin());
	 p != _stochastic_children.end(); ++p) {
      if (ISNAN((*p)->logDensity())) {
	throw NodeError(*p, "Failure to calculate log density");
      }
    }
    throw runtime_error("Error in loglikelihood calculation");
  }
  return logprior + loglikelihood;
}

vector<StochasticNode*> const &Sampler::stochasticChildren() const
{
  return _stochastic_children;
}

vector<DeterministicNode*> const &Sampler::deterministicChildren() const
{
  return _deterministic_children;
}

void Sampler::setValue(vector<double const *> const &value, 
		       vector<long> const &length)
{
  unsigned int n = _nodes.size();
  for (unsigned int i = 0; i < n; ++i) {
     _nodes[i]->data.setValue(value[i], length[i]);
  }
  for (vector<DeterministicNode*>::iterator i(_deterministic_children.begin());
       i != _deterministic_children.end(); ++i) {
    (*i)->forwardSample();
  }
}










