#ifndef SAMPLER_H_
#define SAMPLER_H_

#include <vector>
#include <set>

class StochasticNode;
class DeterministicNode;
class Node;

/**
 * A sampler updates a set of stochastic nodes.  It is also responsible for
 * updating the immediate deterministic descendants of those nodes
 *
 * Some terminology is required to understand what the sampler does
 *
 * The "immediate deterministic descendants" (IDD) of a set of stochastic
 * nodes S are the descendants of S in the graph where all stochastic
 * nodes except those in S have been removed.
 *
 * The "marginal stochastic children" (MSC) of a set of stochastic nodes
 * S are the children of S in the graph where all deterministic nodes
 * have been marginalized out.
 *
 * A node is classified as "informative" if has observed descendants
 * otherwise it is "uninformative".  Uninformative children of a node
 * can be (and generally are) ignored when sampling.
 *
 * A vector of nodes is in "forward sampling order" if node B always
 * appears after node A when there is a path from A to B.
 *
 */
class Sampler {
  std::vector<StochasticNode *> _nodes;
  std::vector<StochasticNode*> _stochastic_children;
  std::vector<DeterministicNode*> _deterministic_children;
public:
  /**
   * Constructs a sampler for the given set of nodes.  The nodes are
   * given as a vector because they must have a well-defined order for
   * the setValue function. Nevertheless, the elements of the vector
   * nodes must be unique.
   */
  Sampler(std::vector<StochasticNode *> const &nodes);
  virtual ~Sampler();
  /**
   * Returns the vector of sampled nodes
   */
  std::vector<StochasticNode *> const &nodes() const;
  /**
   * Sets the values of the sampled nodes.  The informative immediate
   * deterministic descendants are automatically updated.
   *
   * @param value vector of pointers to the beginning of an array of
   * values for each node.  
   * @param length vector of lengths for the corresponding arrays.
   */
  void setValue(std::vector<double const *> const &value, 
		std::vector<long> const &length);
  /**
   * The informative marginal stochastic children of the set
   * of nodes being sampled.
   */
  std::vector<StochasticNode*> const &stochasticChildren() const;
  /**
   * The informative immediate deterministic descendendants of the
   * nodes, in forward sampling order
   */
  std::vector<DeterministicNode*> const &deterministicChildren() const;
  /**
   * Calculates the log conditional density of the sampled nodes,
   * given all other nodes on the graph.  The log full conditional is
   * calculated up to an additive constant.
   */
  double logFullConditional();
  /**
   * Every sampler must update the vector of nodes and its immediate
   * deterministic descendants using the update function.
   */
  virtual void update() = 0;
  /**
   * Static function that classifies the descendants of the given set
   * of stochastic nodes and adds them to one of three vectors.
   * Uninformative children are ignored.
   * @param stoch_nodes Informative marginal stochastic children
   * @param dtrm_nodes Informative immediate deterministic descendants,
   * in forward sampling order
   */
  static void classifyChildren(std::set<StochasticNode const*> const &nodes,
			       std::vector<StochasticNode*> &stoch_nodes,
			       std::vector<DeterministicNode*> &dtrm_nodes);
};

#endif /* SAMPLER_H_ */
