#include <config.h>
#include <sampler/SliceSampler.h>
#include <graph/StochasticNode.h>
#include <graph/NodeError.h>
#include <distributions/DistReal.h>

#include <Rmath.h>

#include <cmath>
#include <cfloat>


SliceSampler::SliceSampler(StochasticNode *node, double width, long nburn,
			   long maxwidth)
  : GibbsSampler(node), _width(width), _nburn(nburn), _sumdiff(0), 
    _iter(0), _maxwidth(maxwidth)
{
}

void SliceSampler::update()
{
  // Test current value
  double g0 = logFullConditional();
  if (!R_finite(g0) && g0 > 0) {
    return;
  }
  if (g0 == -DBL_MAX) {
    throw NodeError(node(), "Current value is inconsistent with data");
  }

  // Generate auxiliary variable
  double z = g0 - rexp(1);

  // Generate random interval of width "_width" about current value
  double xold = node()->data.value()[0];
  double L = xold - unif_rand() * _width; 
  double R = L + _width;

  DistReal const *rdist = asReal(node()->distribution());
  double lower = rdist->lowerLimit(node()->parameters());
  double upper = rdist->upperLimit(node()->parameters());

  // Stepping out 

  // Randomly set number of steps in left and right directions,
  // subject to the limit in the maximal size of the interval
  int j = static_cast<int>(unif_rand() * _maxwidth);
  int k = _maxwidth - 1 - j;


  if (L < lower) {
    L = lower;
  }
  else {
    setValue(&L, 1);
    while (j-- > 0 && logFullConditional() > z) {
      L -= _width;
      if (L < lower) {
	L = lower;
	break;
      }
      setValue(&L, 1);
    }
  }

  if (R > upper) {
    R = upper;
  }
  else {
    setValue(&R, 1);
    while (k-- > 0 && logFullConditional() > z) {
      R += _width;
      if (R > upper) {
	R = upper;
	break;
      }
      setValue(&R, 1);
    }
  }

  // Keep sampling from the interval until acceptance (the loop is
  // guaranteed to terminate).
  double xnew;
  for(;;) {
    xnew =  L + unif_rand() * (R - L);
    setValue(&xnew, 1);
    double g = logFullConditional();
    if (g >= z - DBL_EPSILON) {
      // Accept point
      break;
    }
    else {
      // shrink the interval
      if (xnew < xold) {
	L = xnew;
      }
      else {
	R = xnew;
      }
    }
  }

  if (_iter < _nburn) {
    _sumdiff += fabs(xnew - xold);
    if (_iter == _nburn - 1) {
      // Reset slice width
      _width = _sumdiff/_nburn;
    }
    _iter++;
  }
}

bool SliceSampler::canSample(StochasticNode const *node)
{
  return ((asReal(node->distribution()) != 0) &&
	  (node->data.length() == 1));
}
