#include <config.h>
#include <sarray/Range.h>

#include <algorithm>
#include <stdexcept>

using std::min;
using std::max;
using std::out_of_range;
using std::invalid_argument;
using std::logic_error;

Range::Range()
  : _lower(1), _upper(1)
{
}

Range::Range(Index const &lower, Index const &upper)
  : _lower(lower), _upper(upper)
{
  if (_lower.length() != _upper.length()) {
    throw logic_error("Unequal length upper and lower bounds in Range constructor");
  }
  else {
    int dim = _lower.length();
    for (int i = 0; i < dim; i++) {
      if (_lower[i] < 1) {
	throw out_of_range("invalid lower bound (< 1) in Range constructor");
      }
      if (_upper[i] < _lower[i]) {
	throw out_of_range("upper bound < lower bound in Range constructor");
      }
    }
  }
}

/*
Range::Range(Index const &upper)
  : _lower(upper.length()), _upper(upper)
{
  unsigned int ndim = upper.length();
  for (unsigned int i = 0; i < ndim; i++) {
    if (upper[i] < 1) {
      throw out_of_range("invalid upper bound (< 1) in Range constructor");
    }
  }
}
*/

bool Range::operator==(Range const &other) const
{
  return (_lower == other._lower && _upper == other._upper);
}

long Range::length() const
{
  long length = 1;
  unsigned int ndim = _upper.length();
  for (unsigned int i = 0; i < ndim; i++) {
    length *= (_upper[i] - _lower[i] + 1);
  }
  return length;
}

unsigned int Range::ndim(bool drop) const
{
  unsigned int maxdim = _lower.length();

  if (drop == false) {
    return maxdim;
  }
  else {
    unsigned int size = 0;
    for (unsigned int i = 0; i < maxdim; i++) {
      if (_upper[i] > _lower[i])
	size++;
    }
    return size;
  }
}

Index Range::dim(bool drop) const
{
  int maxdim = _upper.length();
  long *dim = new long[maxdim];
  int size = 0;

  for (int i = 0; i < maxdim; i++) {
    long d = (_upper[i] - _lower[i] + 1);
    if (!drop || d > 1) {
      dim[size] = d;
      size++;
    }
  }
  if (drop && size == 0) {
    delete [] dim;
    return Index(1);
  }
  else {
    Index index(size);
    for (int j = 0; j < size; j++) {
      index[j] = dim[j];
    }
    delete [] dim;
    return index;
  }
}

bool Range::contains(Index const &index) const
{
  unsigned int dim = _upper.length();
  if (index.length() != dim) {
    throw invalid_argument("Dimension mismatch between index and range");
  }
 
  for (unsigned int i = 0; i < dim; i++) {
    if (index[i] < _lower[i] || index[i] > _upper[i])
      return false;
  }
  return true;
}

bool Range::contains(Range const &other) const
{
  return contains(other._lower) && contains(other._upper);
}

bool Range::overlaps(Range const &other) const
{
  unsigned int dim = ndim();
  if (other.ndim() != dim) {
    throw invalid_argument("Dimension mismatch between ranges");
  }

  for (unsigned int i = 0; i < dim; i++) {
    if (max(_lower[i], other._lower[i]) > 
	min(_upper[i], other._upper[i])) {
      return false;
    }
  }
  return true;
}

long Range::leftOffset(Index const &index) const
{
  if (!contains(index)) {
    throw out_of_range("Index outside of allowed range");
  }
 
  long offset = 0;
  long step = 1;
  unsigned int ndim = _upper.length();
  for (unsigned int i = 0; i < ndim; i++) {
    offset += step * (index[i] - _lower[i]);
    step *= (_upper[i] - _lower[i] + 1);
  }
  return offset;
}

long Range::rightOffset(Index const &index) const
{
  if (!contains(index)) {
    throw out_of_range("Index outside of allowed range");
  }
 
  long offset = 0;
  long step = 1;
  for (int i = _upper.length() - 1; i >= 0; --i) {
    offset += step * (index[i] - _lower[i]);
    step *= (_upper[i] - _lower[i] + 1);
  }
  return offset;
}

Index const & Range::lower() const
{
  return _lower;
}

Index const & Range::upper() const
{
  return _upper;
}

int rangeComp(Range const &range1, Range const &range2)
{
  /* 
     Defines a complete ordering among ranges based first
     on lower limit and then on upper limit.
  */
  int ans;

  ans = indexComp(range1.lower(), range2.lower());
  if (ans==0) {
    return indexComp(range1.upper(), range2.upper());
  }
  else {
    return ans;
  }
}


