#ifndef RANGE_H_
#define RANGE_H_

#include <sarray/Index.h>

/**
 * A Range object represents a range of indices used to take a subset
 * of a multi-dimensional array. For example, if A is a 3 x 4 x 5
 * array, then A[1:2,4,4:5] is a subset defined by a range with lower
 * boundary (1,4,4) and upper boundary (2,4,5)
 * @short Range of indices
 * @see Index
 */
class Range {
	Index _lower;
	Index _upper;
 public:
	/**
	 * Default constructor. Produces a one-dimensional range with
	 * lower and upper boundaries equal to 1
	 */
	Range();
	/**
	 * Construct a range based on the upper limit. The lower limit
	 * is assumed to be 1 for each element.
	 * @param upper Index representing upper limits. A range_error
	 *              is thrown if any element of upper is less than 1.
	 * @exception range_error
	*/
	//Range(Index const &upper);
	/**
	 * Construct a range based on given lower and upper limits
	 * A logic_error is thrown if these are of different lengths
	 * @param lower lower limits. A range_error is thrown if any
	 *              of these are < 1
	 * @param upper upper limits. A range_error is thrown if any
	 *              element of upper is smaller than the corresponding
	 *              element of lower.
	 * @exception logic_error range_error
	 */
	Range(Index const &lower, Index const &upper);
	/**
	 * Equality operator
	 */
	bool operator==(Range const &range) const;
	/**
	 * Length of the range. This is the number of indices that
	 * are contained in the range.
	 */
	long length() const;
	/**
	 * Indicates whether the index argument is contained inside the range
	 */
	bool contains(Index const &index) const;
	/**
	 * Indicates whether the range argument is completely contained
	 * inside the range.
	 * @param index Index to test. Must be of correct length or an
	 *              invalid_argument exception is thrown.
	 * @exception invalid_argument
	 */
	bool contains(Range const &range) const;
	/**
	 * Indicates whether two ranges overlap
	 * @param range Range to test. Must have the same number of
	 *               indices or an invalid_argument exception is thrown.
	 */
	bool overlaps(Range const &range) const;
	/**
	 * The number of iterations of a LeftRangeIterator
	 * required to reach index, starting from the lower limit of the range.
	 * @param index Index to convert to offset. An out_of_range is thrown
	 *              if the index is not contained in the range.
	 * @see LeftRangeIterator
	 */
	long leftOffset(Index const &index) const;
	/**
	 * The number of iterations of a RightRangeIterator
	 * required to reach index, starting from the lower limit of the range.
	 * @param index Index to convert to offset. An out_of_range is thrown
	 *              if the index is not contained in the range.
	 * @see RightRangeIterator
	 */
	long rightOffset(Index const &index) const;
	/**
	 * Dimension of the range. The range [1:4,2,3:5] has dimension
	 * (4,1,3) if drop==false and (4,3) if drop==true. 
	 * @param drop Should dimensions of size 1 be dropped?
	 */
	Index dim(bool drop=true) const;
	/**
	 * Number of dimensions covered by the Range
	 * @param drop Should dimensions of size 1 be dropped?
	 */
	unsigned int ndim(bool drop=false) const;

	/**
	 * lower limit of range
	 */
	Index const & lower() const;
	/**
	 * upper limit of range
	 */
	Index const & upper() const;
};

int rangeComp(Range const &range1, Range const &range2);

struct ltrange
{
	/* "Less than" operator for Range */
	bool operator()(Range const &r1, Range const &r2) const
	{
		return (rangeComp(r1, r2) < 0);
	}
};

#endif /* RANGE_H_ */
